from typing import Final

from .exc import *  # noqa: F403
from .format import format_invoice_as_text as format_invoice_as_text
from .generate import (
    generate_et as generate_et,
    generate_xml as generate_xml,
)
from .model import *  # noqa: F403
from .money import Money as Money
from .parse import parse_xml as parse_xml
from .pdf_common import FileRelationship as FileRelationship
from .pdf_embed import (
    embed_facturx_file_in_pdf as embed_facturx_file_in_pdf,
    embed_invoice_in_pdf as embed_invoice_in_pdf,
)
from .pdf_parse import parse_pdf as parse_pdf

FACTURX_VERSION: Final = "1.0.07"
ZUGFERD_VERSION: Final = "2.3"



from enum import Enum
from typing import Final

FACTURX_FILENAME: Final = "factur-x.xml"
XRECHNUNG_FILENAME: Final = "xrechnung.xml"
FACTURX_XML_VERSION: Final = "1.0"


class FileRelationship(Enum):
    """The relationship between a PDF file and an embedded file.

    * DATA: The embedded file contains data that is displayed in the PDF file.
    * SOURCE: The embedded file is the source data for the PDF file.
    * ALTERNATIVE: The embedded file is an alternative representation of
      the PDF file.
    * SUPPLEMENT: The embedded file contains information that supplements
      the data in the PDF file.
    """

    DATA = "Data"
    SOURCE = "Source"
    ALTERNATIVE = "Alternative"
    SUPPLEMENT = "Supplement"



from __future__ import annotations

import sys
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Final
from xml.dom.minidom import parseString

from pypdf import PdfWriter
from pypdf.generic import (
    ArrayObject,
    DictionaryObject,
    NameObject,
    create_string_object,
)
from pypdf.xmp import XmpInformation

from .exc import InsufficientPDFError
from .generate import generate_xml
from .model import BasicInvoice, MinimumInvoice
from .pdf_common import FACTURX_FILENAME, FACTURX_XML_VERSION, FileRelationship
from .types import Profile

if TYPE_CHECKING:
    from _typeshed import StrPath

RDF_NS: Final = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
FACTURX_NS: Final = "urn:factur-x:pdfa:CrossIndustryDocument:invoice:1p0#"

FX_EXTENSION_SCHEMA: Final = """
    <rdf:Description xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:pdfaExtension="http://www.aiim.org/pdfa/ns/extension/" xmlns:pdfaSchema="http://www.aiim.org/pdfa/ns/schema#" xmlns:pdfaProperty="http://www.aiim.org/pdfa/ns/property#" rdf:about="">
      <pdfaExtension:schemas>
        <rdf:Bag>
          <rdf:li rdf:parseType="Resource">
            <pdfaSchema:schema>Factur-X PDFA Extension Schema</pdfaSchema:schema>
            <pdfaSchema:namespaceURI>urn:factur-x:pdfa:CrossIndustryDocument:invoice:1p0#</pdfaSchema:namespaceURI>
            <pdfaSchema:prefix>fx</pdfaSchema:prefix>
            <pdfaSchema:property>
              <rdf:Seq>
                <rdf:li rdf:parseType="Resource">
                  <pdfaProperty:name>DocumentFileName</pdfaProperty:name>
                  <pdfaProperty:valueType>Text</pdfaProperty:valueType>
                  <pdfaProperty:category>external</pdfaProperty:category>
                  <pdfaProperty:description>The name of the embedded XML document</pdfaProperty:description>
                </rdf:li>
                <rdf:li rdf:parseType="Resource">
                  <pdfaProperty:name>DocumentType</pdfaProperty:name>
                  <pdfaProperty:valueType>Text</pdfaProperty:valueType>
                  <pdfaProperty:category>external</pdfaProperty:category>
                  <pdfaProperty:description>The type of the hybrid document in capital letters, e.g. INVOICE or ORDER</pdfaProperty:description>
                </rdf:li>
                <rdf:li rdf:parseType="Resource">
                  <pdfaProperty:name>Version</pdfaProperty:name>
                  <pdfaProperty:valueType>Text</pdfaProperty:valueType>
                  <pdfaProperty:category>external</pdfaProperty:category>
                  <pdfaProperty:description>The actual version of the standard applying to the embedded XML document</pdfaProperty:description>
                </rdf:li>
                <rdf:li rdf:parseType="Resource">
                  <pdfaProperty:name>ConformanceLevel</pdfaProperty:name>
                  <pdfaProperty:valueType>Text</pdfaProperty:valueType>
                  <pdfaProperty:category>external</pdfaProperty:category>
                  <pdfaProperty:description>The conformance level of the embedded XML document</pdfaProperty:description>
                </rdf:li>
              </rdf:Seq>
            </pdfaSchema:property>
          </rdf:li>
        </rdf:Bag>
      </pdfaExtension:schemas>
    </rdf:Description>
"""  # noqa: E501


def embed_facturx_file_in_pdf(
    pdf_filename: str | Path,
    xml_filename: StrPath,
    *,
    profile: Profile,
    relationship: FileRelationship = FileRelationship.DATA,
) -> bytes:
    """Embed a Factur-X XML file into a PDF file.

    The input PDF file must already be a valid PDF/A-3 document, otherwise
    the generated PDF won't be a valid Factur-X PDF.

    Returns the modified PDF file as a byte stream.

    The `relationship` parameter specifies the relationship of the
    embedded file to the PDF file. It can be one of the values from
    the `FileRelationship` enum.
    """

    xml_data = Path(xml_filename).read_text(encoding="utf-8")
    return _embed(pdf_filename, xml_data, profile, relationship=relationship)


def embed_invoice_in_pdf(
    pdf_filename: str | Path,
    invoice: MinimumInvoice,
    *,
    relationship: FileRelationship | None = None,
) -> bytes:
    """Embed a Factur-X invoice into a PDF file.

    The input PDF file must already be a valid PDF/A-3 document, otherwise
    the generated PDF won't be a valid Factur-X PDF.

    Returns the modified PDF file as a byte stream.

    The `relationship` parameter specifies the relationship of the
    embedded file to the PDF file. It can be one of the values from
    the `FileRelationship` enum. By default, the `Data` relationship
    is used for MINIMUM and BASIC WL profiles, while the `Alternative`
    relationship is used for all other profiles.
    """

    if relationship is None:
        if isinstance(invoice, BasicInvoice):
            relationship = FileRelationship.ALTERNATIVE
        else:
            relationship = FileRelationship.DATA

    xml_data = generate_xml(invoice)

    return _embed(
        pdf_filename, xml_data, invoice.PROFILE_NAME, relationship=relationship
    )


def _embed(
    pdf_filename: str | Path,
    xml_data: str,
    profile: Profile,
    relationship: FileRelationship,
) -> bytes:
    writer = PdfWriter(clone_from=pdf_filename)
    _set_metadata(writer, profile)
    _add_attachment(writer, xml_data, relationship)

    output = BytesIO()
    writer.write_stream(output)
    return output.getvalue()


def _set_metadata(writer: PdfWriter, profile: Profile) -> None:
    # TODO: We can't use writer.xmp_metadata, because that will replace
    # an indirect /Metadata entry in the PDF with a new one.
    meta_o = writer.root_object.get("/Metadata")
    if meta_o is None:
        raise InsufficientPDFError("Missing XMP metadata in PDF file.")
    meta_o = meta_o.get_object()
    meta = XmpInformation(meta_o)

    schema = parseString(FX_EXTENSION_SCHEMA)
    assert schema.documentElement is not None
    del schema.documentElement.attributes["xmlns:rdf"]

    doc = meta.rdf_root.ownerDocument
    assert doc is not None
    meta.rdf_root.appendChild(schema.documentElement)
    root = doc.createElementNS(RDF_NS, "rdf:Description")
    root.setAttribute("rdf:about", "")
    meta.rdf_root.appendChild(root)

    root.attributes["xmlns:fx"] = FACTURX_NS
    for tag, content in [
        ("DocumentType", "INVOICE"),
        ("DocumentFileName", FACTURX_FILENAME),
        ("Version", FACTURX_XML_VERSION),
        ("ConformanceLevel", profile),
    ]:
        el = doc.createElement(f"fx:{tag}")
        el.appendChild(doc.createTextNode(content))
        root.appendChild(el)
    meta_o.set_data(doc.toxml().encode("utf-8"))


def _add_attachment(
    writer: PdfWriter, xml_data: str, relationship: FileRelationship
) -> None:
    writer.add_attachment(
        filename=FACTURX_FILENAME, data=xml_data.encode("utf-8")
    )
    attachment = list(writer.attachment_list)[-1]
    attachment.pdf_object[NameObject("/UF")] = create_string_object(
        FACTURX_FILENAME
    )
    attachment.pdf_object[NameObject("/AFRelationship")] = NameObject(
        f"/{relationship.value}"
    )
    attachment._embedded_file[NameObject("/Subtype")] = NameObject("/text/xml")

    # Replace file dictionary with an indirect object.
    catalog = writer._root_object
    names_dict = catalog["/Names"]
    assert isinstance(names_dict, DictionaryObject)
    embedded_files = names_dict["/EmbeddedFiles"]
    assert isinstance(embedded_files, DictionaryObject)
    names_array = embedded_files["/Names"]
    assert isinstance(names_array, ArrayObject)
    file_dict = names_array[-1]
    assert isinstance(file_dict, DictionaryObject)
    file_ref = writer._add_object(file_dict)
    names_array[-1] = file_ref

    # Add the file reference to the /AF array in the catalog.
    if NameObject("/AF") not in catalog:
        catalog[NameObject("/AF")] = ArrayObject()
    af_array = catalog[NameObject("/AF")]
    assert isinstance(af_array, ArrayObject)
    af_array.append(file_ref)


def main() -> None:
    from ._test_data import TEST_EN16931_INVOICE
    from .parse import parse_xml

    if len(sys.argv) < 2 or len(sys.argv) > 3:
        print(f"Usage: {sys.argv[0]} PDF-FILE [XML-FILE]", file=sys.stderr)
        sys.exit(1)
    pdf_filename = sys.argv[1]
    if len(sys.argv) == 2:
        pdf_data = embed_invoice_in_pdf(pdf_filename, TEST_EN16931_INVOICE)
    else:
        invoice = parse_xml(sys.argv[2])
        pdf_data = embed_facturx_file_in_pdf(
            pdf_filename, sys.argv[2], profile=invoice.PROFILE_NAME
        )
    with open("embedded.pdf", "xb") as f:
        f.write(pdf_data)


if __name__ == "__main__":
    main()



from __future__ import annotations

import datetime
from collections.abc import Sequence
from dataclasses import KW_ONLY, dataclass, field
from decimal import Decimal
from typing import ClassVar, Literal

from .const import (
    ALLOWED_ATTACHMENT_MIME_TYPES,
    URN_BASIC_PROFILE,
    URN_BASIC_WL_PROFILE,
    URN_EN16931_PROFILE,
    URN_MINIMUM_PROFILE,
)
from .countries import validate_iso_3166_1_alpha_2
from .exc import ModelError
from .money import Money, validate_iso_4217_currency
from .type_codes import (
    AllowanceChargeCode,
    DocumentTypeCode,
    ItemTypeCode,
    PaymentMeansCode,
    PaymentTimeCode,
    ReferenceQualifierCode,
    SpecialServiceCode,
    TaxCategoryCode,
    TextSubjectCode,
    VATExemptionCode,
)
from .types import ID, Attachment, DocRef, OptionalQuantity, Profile, Quantity

__all__ = [
    "MinimumInvoice",
    "BasicWLInvoice",
    "BasicInvoice",
    "EN16931Invoice",
    "LineItem",
    "EN16931LineItem",
    "PostalAddress",
    "ReferenceDocument",
    "Tax",
    "TradeContact",
    "TradeParty",
    "LineAllowance",
    "LineCharge",
    "DocumentAllowance",
    "DocumentCharge",
    "ProductCharacteristic",
    "ProductClassification",
    "PaymentMeans",
    "PaymentTerms",
    "BankAccount",
    "IncludedNote",
]


@dataclass
class IncludedNote:
    """A note included in an invoice."""

    content: str
    subject_code: TextSubjectCode | None = None


@dataclass
class TradeParty:
    """
    Trade party data used in invoices for seller, buyer, and other parties.

    The required fields depend on the invoice profile and the role of the
    party.
    """

    name: str | None
    address: PostalAddress | None  # Not required in MINIMUM for the buyer
    email: str | None = None
    _: KW_ONLY
    tax_number: str | None = None
    vat_id: str | None = None
    ids: Sequence[str] = field(default_factory=list)
    global_ids: Sequence[ID] = field(default_factory=list)
    description: str | None = None
    legal_id: ID | None = None
    trading_business_name: str | None = None
    contact: TradeContact | None = None

    def validate(
        self,
        profile: type[MinimumInvoice],
        *,
        which: Literal[
            "seller", "buyer", "seller tax representative", "ship to", "payee"
        ],
        has_representative: bool = False,
    ) -> None:
        """Validate the requirements for the given profile."""

        if which in ("seller tax representative", "ship to", "payee"):
            if not issubclass(profile, BasicWLInvoice):
                raise ModelError(
                    f"{which.capitalize} is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )

        # ids and global_ids

        for gid in self.global_ids:
            if gid[1] is None:
                raise ModelError("Global ID scheme ID is required.")

        if which in ("buyer", "ship to", "payee"):
            if issubclass(profile, BasicWLInvoice):
                if len(self.ids) > 1:
                    raise ModelError(
                        f"Multiple {which} IDs are not allowed in the "
                        f"{profile.PROFILE_NAME} profile."
                    )
                if len(self.global_ids) > 1:
                    raise ModelError(
                        f"Multiple {which} global IDs are not allowed in the "
                        f"{profile.PROFILE_NAME} profile."
                    )
            else:  # only relevant for "buyer"
                if len(self.ids) > 0:
                    raise ModelError(
                        f"{which.capitalize()} IDs are not allowed in the "
                        f"{profile.PROFILE_NAME} profile."
                    )
                if len(self.global_ids) > 0:
                    raise ModelError(
                        f"{which.capitalize()} global IDs are not allowed "
                        f"in the {profile.PROFILE_NAME} profile."
                    )

        elif which == "seller":
            if (
                self.vat_id is None
                and self.tax_number is None
                and not has_representative
            ):
                raise ModelError(
                    "Seller must have a VAT ID or tax number in the "
                    f"{profile.PROFILE_NAME} profile."
                )
            if not issubclass(profile, BasicWLInvoice):
                if len(self.ids) > 0:
                    raise ModelError(
                        "Seller IDs are not allowed in the "
                        f"{profile.PROFILE_NAME} profile."
                    )
                if len(self.global_ids) > 0:
                    raise ModelError(
                        "Seller global IDs are not allowed in the "
                        f"{profile.PROFILE_NAME} profile."
                    )
        elif which == "seller tax representative":
            if len(self.ids) > 0:
                raise ModelError(
                    "Seller tax representative IDs are not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )
            if len(self.global_ids) > 0:
                raise ModelError(
                    "Seller tax representative global IDs are not allowed "
                    f"in the {profile.PROFILE_NAME} profile."
                )

        # name

        # Optional for ship to party only.
        if which != "ship to" and self.name is None:
            raise ModelError(f"{which.capitalize()} name is required.")

        # description

        # Allowed only for seller in EN 16931/COMFORT.
        if which != "seller" or not issubclass(profile, EN16931Invoice):
            if self.description is not None:
                raise ModelError(
                    f"{which.capitalize()} description is not allowed."
                )

        # legal organization ID

        # Not allowed for seller tax representative and ship to party.
        if which in ("seller tax representative", "ship to"):
            if self.legal_id is not None:
                raise ModelError(
                    f"{which.capitalize()} legal ID is not allowed "
                    f"in the {profile.PROFILE_NAME} profile."
                )

        # legal organization trading business name

        # Allowed for buyer in EN 16931/COMFORT and for seller in BASIC WL.
        if which == "buyer":
            if not issubclass(profile, EN16931Invoice):
                if self.trading_business_name is not None:
                    raise ModelError(
                        "Buyer trading business name is not allowed "
                        f"in the {profile.PROFILE_NAME} profile."
                    )
        elif which == "seller":
            if not issubclass(profile, BasicWLInvoice):
                if self.trading_business_name is not None:
                    raise ModelError(
                        "Seller trading business name is not allowed in the "
                        f"{profile.PROFILE_NAME} profile."
                    )
        else:
            if self.trading_business_name is not None:
                raise ModelError(
                    f"{which.capitalize()} trading business name is not "
                    f"allowed in the {profile.PROFILE_NAME} profile."
                )

        # contacts

        # Only allowed for buyer and seller in EN 16931/COMFORT.
        if which not in ("buyer", "seller") or not issubclass(
            profile, EN16931Invoice
        ):
            if self.contact is not None:
                raise ModelError(
                    f"{which.capitalize()} contacts are not allowed "
                    f"in the {profile.PROFILE_NAME} profile."
                )

        # address

        # Not allowed for payee.
        # Required for seller, seller tax representative in all profiles,
        # and buyer in BASIC WL.
        # Optional for buyer in MINIMUM and for ship to party.
        if which == "payee":
            if self.address is not None:
                raise ModelError(
                    "Payee address is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )
        elif which in ("seller", "seller tax representative") or (
            which == "buyer" and issubclass(profile, BasicWLInvoice)
        ):
            if self.address is None:
                raise ModelError(
                    f"{which.capitalize()} address is required in the "
                    f"{profile.PROFILE_NAME} profile."
                )

        if self.address is not None:
            self.address.validate(profile)

        # email

        # Not allowed in MINIMUM.
        if not issubclass(profile, BasicWLInvoice):
            if self.email is not None:
                raise ModelError(
                    f"{which.capitalize()} email is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )

        # tax number

        # Not allowed for buyer in MINIMUM and other parties.
        # Optional buyer in BASIC WL and seller.
        if (
            which == "buyer" and not issubclass(profile, BasicWLInvoice)
        ) or which in ("seller tax representative", "ship to", "payee"):
            if self.tax_number is not None:
                raise ModelError(
                    f"{which.capitalize()} tax number is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )

        # VAT ID

        # Required for seller tax representative.
        # Not allowed for buyer in MINIMUM and other parties.
        # Optional for buyer in BASIC WL and seller.
        if which == "seller tax representative":
            if self.vat_id is None:
                raise ModelError(
                    "Seller tax representative VAT ID is required in the "
                    f"{profile.PROFILE_NAME} profile."
                )
        elif (
            which == "buyer"
            and not issubclass(profile, BasicWLInvoice)
            or which in ("ship to", "payee")
        ):
            if self.vat_id is not None:
                raise ModelError(
                    f"{which.capitalize()} VAT ID is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )


@dataclass
class TradeContact:
    """Contact information for a trade party."""

    person_name: str | None = None
    department_name: str | None = None
    _: KW_ONLY
    phone: str | None = None
    email: str | None = None


@dataclass
class PostalAddress:
    """Postal address used in invoices."""

    country_code: str  # ISO 3166-1 alpha-2
    country_subdivision: str | None = None
    post_code: str | None = None
    city: str | None = None
    line_one: str | None = None
    line_two: str | None = None
    line_three: str | None = None

    def __post_init__(self) -> None:
        if not validate_iso_3166_1_alpha_2(self.country_code):
            raise ModelError("Invalid ISO 3166-1 alpha-2 country code.")

    def validate(self, profile: type[MinimumInvoice]) -> None:
        """Validate the requirements for the given profile."""
        if not issubclass(profile, BasicWLInvoice):
            if (
                self.country_subdivision is not None
                or self.post_code is not None
                or self.city is not None
                or self.line_one is not None
                or self.line_two is not None
                or self.line_three is not None
            ):
                raise ModelError(
                    "Address fields are not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )


@dataclass
class MinimumInvoice:
    """Invoice data for the MINIMUM profile."""

    PROFILE_NAME: ClassVar[Profile] = "MINIMUM"
    PROFILE_URN: ClassVar[str] = URN_MINIMUM_PROFILE

    invoice_number: str
    type_code: DocumentTypeCode
    invoice_date: datetime.date
    seller: TradeParty
    buyer: TradeParty
    currency_code: str

    _: KW_ONLY

    line_total_amount: Money | None = None
    tax_basis_total_amount: Money
    tax_total_amounts: list[Money]
    grand_total_amount: Money
    due_payable_amount: Money

    business_process_id: str | None = None
    buyer_reference: str | None = None
    buyer_order_id: str | None = None

    def __post_init__(self) -> None:
        if not self.type_code.is_invoice_type:
            raise ModelError(f"Invalid invoice type code: {self.type_code}.")
        self.seller.validate(
            type(self),
            which="seller",
            has_representative=isinstance(self, BasicWLInvoice)
            and self.seller_tax_representative is not None,
        )
        self.buyer.validate(type(self), which="buyer")
        validate_iso_4217_currency(self.currency_code)
        if type(self) is MinimumInvoice:
            if len(self.tax_total_amounts) > 1:
                raise ModelError(
                    "Multiple tax total amounts are not allowed in the "
                    "MINIMUM profile."
                )


@dataclass
class BasicWLInvoice(MinimumInvoice):
    """Invoice data for the BASIC WL profile."""

    PROFILE_NAME: ClassVar[Profile] = "BASIC WL"
    PROFILE_URN = URN_BASIC_WL_PROFILE

    tax: Sequence[Tax]

    _: KW_ONLY

    charge_total_amount: Money | None = None
    allowance_total_amount: Money | None = None
    prepaid_amount: Money | None = None
    payee: TradeParty | None = None
    delivery_date: datetime.date | None = None
    billing_period: tuple[datetime.date, datetime.date] | None = None
    allowances: Sequence[DocumentAllowance] = field(default_factory=list)
    charges: Sequence[DocumentCharge] = field(default_factory=list)
    notes: list[IncludedNote] = field(default_factory=list)
    seller_tax_representative: TradeParty | None = None
    contract_id: str | None = None
    ship_to: TradeParty | None = None
    despatch_advice_id: str | None = None
    sepa_reference: str | None = None
    payment_reference: str | None = None
    payment_means: Sequence[PaymentMeans] = field(default_factory=list)
    payment_terms: PaymentTerms | None = None
    preceding_invoices: list[tuple[str, datetime.date | None]] = field(
        default_factory=list
    )
    receiver_accounting_ids: Sequence[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.line_total_amount is None:
            raise ModelError(
                "Line total amount is required in BASIC WL profile."
            )
        if len(self.tax) < 1:
            raise ModelError("At least one tax entry is required.")
        for tax in self.tax:
            tax.validate(type(self))
        if self.payee is not None:
            self.payee.validate(type(self), which="payee")
        if self.seller_tax_representative is None:
            pass
        else:
            self.seller_tax_representative.validate(
                type(self), which="seller tax representative"
            )
        if self.ship_to is not None:
            self.ship_to.validate(type(self), which="ship to")
        for means in self.payment_means:
            means.validate(type(self))
        if self.payment_terms is not None:
            self.payment_terms.validate(type(self))
        if len(self.tax_total_amounts) > 2:
            raise ModelError(
                "Multiple tax total amounts are not allowed in the "
                f"{self.PROFILE_NAME} profile."
            )


@dataclass
class BasicInvoice(BasicWLInvoice):
    """Invoice data for the BASIC profile."""

    PROFILE_NAME: ClassVar[Profile] = "BASIC"
    PROFILE_URN = URN_BASIC_PROFILE

    _: KW_ONLY

    line_items: Sequence[LineItem]  # BG-25

    def __post_init__(self) -> None:
        if len(self.line_items) < 1:
            raise ModelError("At least one line item is required.")
        if type(self) is BasicInvoice:
            for li in self.line_items:
                if isinstance(li, EN16931LineItem):
                    raise TypeError(
                        "EN 16931/COMFORT line items are not allowed in the "
                        "BASIC profile."
                    )
        for li in self.line_items:
            li.validate(type(self))
        if len(self.receiver_accounting_ids) > 1:
            raise ModelError(
                "Multiple accounting reference IDs are not allowed in the "
                f"{self.PROFILE_NAME} profile."
            )


@dataclass
class EN16931Invoice(BasicInvoice):
    """Invoice data for the EN 16931/COMFORT profile."""

    PROFILE_NAME: ClassVar[Profile] = "EN 16931"
    PROFILE_URN = URN_EN16931_PROFILE

    _: KW_ONLY

    receiving_advice_id: str | None = None
    rounding_amount: Money | None = None
    seller_order_id: str | None = None
    referenced_docs: Sequence[ReferenceDocument] = field(default_factory=list)
    procuring_project: tuple[str, str] | None = None
    tax_currency_code: str | None = None


@dataclass
class LineItem:
    """Line item data used in the BASIC profile."""

    id: str
    name: str
    net_price: Money
    billed_quantity: Quantity
    billed_total: Money
    tax_rate: Decimal | None
    tax_category: TaxCategoryCode = TaxCategoryCode.STANDARD_RATE
    global_id: ID | None = None

    _: KW_ONLY
    basis_quantity: OptionalQuantity | None = None
    allowances: Sequence[LineAllowance] = field(default_factory=list)
    charges: Sequence[LineCharge] = field(default_factory=list)

    def validate(self, profile: type[BasicInvoice]) -> None:
        """Validate the requirements for the given profile."""
        for allowance in self.allowances:
            allowance.validate(profile)
        for charge in self.charges:
            charge.validate(profile)


@dataclass
class EN16931LineItem(LineItem):
    """Line item data used in the EN 16931/COMFORT profile."""

    _: KW_ONLY
    description: str | None = None
    note: IncludedNote | None = None
    # (Unit price, optional basis quantity)
    gross_unit_price: tuple[Money, OptionalQuantity | None] | None = None
    gross_allowance_or_charge: LineAllowance | LineCharge | None = None
    seller_assigned_id: str | None = None
    buyer_assigned_id: str | None = None
    product_characteristics: Sequence[ProductCharacteristic] = field(
        default_factory=list
    )
    product_classifications: Sequence[ProductClassification] = field(
        default_factory=list
    )
    origin_country: str | None = None
    buyer_order_line_id: str | None = None
    billing_period: tuple[datetime.date, datetime.date] | None = None
    doc_ref: DocRef | None = None
    trade_account_id: str | None = None

    def __post_init__(self) -> None:
        if self.note is not None and self.note.subject_code is not None:
            raise ModelError(
                "Line item note subject codes are not allowed in the "
                "EN 16931/COMFORT profile."
            )
        if (
            self.gross_allowance_or_charge is not None
            and self.gross_unit_price is None
        ):
            raise ModelError(
                "Allowance/charge requires a gross unit price in the "
                "EN 16931/COMFORT profile."
            )
        if self.origin_country is not None:
            if not validate_iso_3166_1_alpha_2(self.origin_country):
                raise ModelError("Invalid ISO 3166-1 alpha-2 country code.")
        if self.billing_period is not None:
            start, end = self.billing_period
            if start > end:
                raise ModelError(
                    "Billing period start date must be before end date."
                )


@dataclass
class ProductCharacteristic:
    """A single product characteristic."""

    description: str
    value: str


@dataclass
class ProductClassification:
    """A single product classification."""

    class_code: str
    _: KW_ONLY
    list_id: ItemTypeCode
    list_version_id: str | None = None


@dataclass
class LineAllowance:
    """An allowance for a line item."""

    actual_amount: Money
    reason_code: AllowanceChargeCode | None = None
    reason: str | None = None
    _: KW_ONLY
    percent: Decimal | None = None
    basis_amount: Money | None = None

    def validate(self, profile: type[BasicInvoice]) -> None:
        """Validate the requirements for the given profile."""
        if not issubclass(profile, EN16931Invoice):
            if self.percent is not None:
                raise ModelError(
                    "Percentage-based allowances are not allowed "
                    f"in the {profile.PROFILE_NAME} profile."
                )
            if self.basis_amount is not None:
                raise ModelError(
                    "Basis amount-based allowances are not allowed "
                    f"in the {profile.PROFILE_NAME} profile."
                )


@dataclass
class LineCharge:
    """A surcharge for a line item."""

    actual_amount: Money
    reason_code: SpecialServiceCode | None = None
    reason: str | None = None
    _: KW_ONLY
    percent: Decimal | None = None
    basis_amount: Money | None = None

    def validate(self, profile: type[BasicInvoice]) -> None:
        """Validate the requirements for the given profile."""

        if self.reason_code is None and not self.reason:
            raise ModelError(
                "Line charge must have a reason code or reason text."
            )

        if not issubclass(profile, EN16931Invoice):
            if self.percent is not None:
                raise ModelError(
                    "Percentage-based charges are not allowed "
                    f"in the {profile.PROFILE_NAME} profile."
                )
            if self.basis_amount is not None:
                raise ModelError(
                    "Basis amount-based charges are not allowed "
                    f"in the {profile.PROFILE_NAME} profile."
                )


@dataclass
class DocumentAllowance(LineAllowance):
    """An allowance for the entire invoice."""

    _: KW_ONLY
    tax_category: TaxCategoryCode = TaxCategoryCode.STANDARD_RATE
    tax_rate: Decimal | None = None


@dataclass
class DocumentCharge(LineCharge):
    """A surcharge for the entire invoice."""

    _: KW_ONLY
    tax_category: TaxCategoryCode = TaxCategoryCode.STANDARD_RATE
    tax_rate: Decimal | None = None


@dataclass
class ReferenceDocument:
    """A reference document attached to an invoice."""

    id: str
    type_code: DocumentTypeCode
    name: str | None = None
    url: str | None = None
    attachment: Attachment | None = None
    _: KW_ONLY
    reference_type_code: ReferenceQualifierCode | None = None

    def __post_init__(self) -> None:
        if not self.type_code.is_supporting_document_type:
            raise ModelError(
                f"Invalid reference document type code: {self.type_code}."
            )
        if (
            self.attachment
            and self.attachment[1] not in ALLOWED_ATTACHMENT_MIME_TYPES
        ):
            raise ModelError(
                f"Unsupported attachment MIME type: {self.attachment[1]}."
            )


@dataclass
class PaymentMeans:
    """Payment means data used in invoices."""

    type_code: PaymentMeansCode
    payee_account: BankAccount | None = None
    payee_bic: str | None = None
    information: str | None = None
    _: KW_ONLY
    # (account number (PAN), cardholder name)
    card: tuple[str, str | None] | None = None
    payer_iban: str | None = None

    def validate(self, profile: type[BasicWLInvoice]) -> None:
        """Validate the requirements for the given profile."""
        if not issubclass(profile, EN16931Invoice):
            if self.information is not None:
                raise ModelError(
                    "Payment means information is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )
            if self.card is not None:
                raise ModelError(
                    "Payment means card information is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )
            if (
                self.payee_account is not None
                and self.payee_account.name is not None
            ):
                raise ModelError(
                    "Payment means account name is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )
            if self.payee_bic is not None:
                raise ModelError(
                    "Payment means BIC is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )


@dataclass
class PaymentTerms:
    """Payment terms data used in invoices."""

    _: KW_ONLY
    description: str | None = None  # EN16931+
    due_date: datetime.date | None = None
    direct_debit_mandate_id: str | None = None

    def validate(self, profile: type[BasicWLInvoice]) -> None:
        """Validate the requirements for the given profile."""
        if not issubclass(profile, EN16931Invoice):
            if self.description is not None:
                raise ModelError(
                    "Payment terms description is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )


@dataclass
class BankAccount:
    """Bank account data used in invoices."""

    iban: str | None
    name: str | None
    bank_id: str | None


@dataclass
class Tax:
    """A single tax entry for an invoice."""

    calculated_amount: Money
    basis_amount: Money
    rate_percent: Decimal | None
    category_code: TaxCategoryCode = TaxCategoryCode.STANDARD_RATE
    _: KW_ONLY
    exemption_reason: str | None = None
    exemption_reason_code: VATExemptionCode | None = None
    tax_point_date: datetime.date | None = None
    due_date_type_code: PaymentTimeCode | None = None

    def __post_init__(self) -> None:
        if (
            self.due_date_type_code is not None
            and not self.due_date_type_code.is_invoice_due_date
        ):
            raise ModelError(
                "Invalid due date type code: {self.due_date_type_code}."
            )
        if (
            self.rate_percent is not None
            and self.basis_amount * self.rate_percent / Decimal(100)
            != self.calculated_amount
        ):
            raise ModelError(
                f"Calculated amount {self.calculated_amount} does not match "
                f"basis amount {self.basis_amount} and tax rate "
                f"{self.rate_percent}\u2009%."
            )

    def validate(self, profile: type[BasicWLInvoice]) -> None:
        """Validate the requirements for the given profile."""
        if not issubclass(profile, EN16931Invoice):
            if self.tax_point_date is not None:
                raise ModelError(
                    "Tax point date is not allowed in the "
                    f"{profile.PROFILE_NAME} profile."
                )



import gettext
from collections.abc import Callable
from pathlib import Path

LOCALE_PATH = Path(__file__).parent / "locale"


def setup_locale() -> Callable[[str], str]:
    """Setup the gettext translation function in a module.

    Usage:

    from ._locale import setup_locale
    _ = setup_locale()

    print(_("Hello, World!"))
    """

    t = gettext.translation("pycheval", localedir=LOCALE_PATH, fallback=True)
    return t.gettext



"""Common types for PyCheval."""

from decimal import Decimal
from typing import Literal, TypeAlias

from .quantities import QuantityCode
from .type_codes import IdentifierSchemeCode, ReferenceQualifierCode

# (ID, scheme ID)
# The scheme ID must be from the ISO/IEC 6523 list.
ID: TypeAlias = tuple[str, IdentifierSchemeCode | str | None]
Quantity: TypeAlias = tuple[Decimal, QuantityCode]
OptionalQuantity: TypeAlias = tuple[Decimal, QuantityCode | None]
# (content, mime type, filename)
Attachment: TypeAlias = tuple[bytes, str, str]
# (issuer assigned ID, reference type code)
DocRef: TypeAlias = tuple[str | None, ReferenceQualifierCode | None]
Profile: TypeAlias = Literal[
    "MINIMUM", "BASIC WL", "BASIC", "EN 16931", "EXTENDED", "XRECHNUNG"
]



class FacturXError(Exception):
    """Base class for Factur-X exceptions."""


class PDFError(FacturXError):
    """Base class for PDF processing exceptions."""


class InsufficientPDFError(PDFError):
    """Raised when a PDF file is insufficient for Factur-X."""


class PDFParseError(FacturXError):
    """Raise when a PDF file cannot be processed."""


class NoFacturXError(PDFParseError):
    """Raised when a PDF file does not contain a Factur-X XML file."""


class ModelError(FacturXError):
    """Raised when a Factur-X model is invalid."""


class FacturXParseError(FacturXError):
    """Base class for Factur-X parsing exceptions."""


class XMLParseError(FacturXParseError):
    """Raised when an XML file cannot be parsed."""


class NotFacturXError(FacturXParseError):
    """Raised when an XML file is not a Factur-X file."""


class UnsupportedProfileError(FacturXParseError):
    """Raised when a Factur-X file uses an unsupported profile."""


class InvalidXMLError(FacturXParseError):
    """Raised when a Factur-X XML file has the wrong structure."""


class InvalidProfileError(FacturXParseError):
    """Raised when a Factur-X file is invalid for the specified profile."""

    def __init__(self, profile_name: str, message: str) -> None:
        super().__init__(message)
        self.profile_name = profile_name



import datetime
from decimal import Decimal
from typing import Final

from .model import (
    EN16931Invoice,
    EN16931LineItem,
    IncludedNote,
    LineAllowance,
    LineCharge,
    LineItem,
    PaymentMeans,
    PaymentTerms,
    PostalAddress,
    ProductCharacteristic,
    ProductClassification,
    ReferenceDocument,
    Tax,
    TradeContact,
    TradeParty,
)
from .money import Money
from .quantities import QuantityCode
from .type_codes import (
    AllowanceChargeCode,
    DocumentTypeCode,
    ItemTypeCode,
    PaymentMeansCode,
    ReferenceQualifierCode,
    TaxCategoryCode,
    TextSubjectCode,
)

TEST_PNG = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\n\x00\x00\x00\n\x08\x02\x00\x00\x00\x02PX\xea\x00\x00\x00&IDAT\x18\xd3c\xfc\xff\xff?\x03n\xc0\xc4\xc0\xc0\xc0\xc0\xc8\x88E\x86\x91\x91\x81\x81\x81\x91\x08\xdd\x83X\x9al\x8f\x01\x00\xdf\x94\t\x11%\xfb\xb4\xe3\x00\x00\x00\x00IEND\xaeB`\x82"  # noqa: E501

TEST_SELLER: Final = TradeParty(
    "Seller GmbH",
    PostalAddress(
        "DE", None, "44331", "Test City", "Test Street 42", None, None
    ),
    "seller@example.com",
    vat_id="DE123456789",
    tax_number="123/456/789",
    ids=["SELLER-123"],
    global_ids=[("123456789", "0088")],
    description="being formed",
    legal_id=("DE123456789", "0088"),
    trading_business_name="Great Business",
    contact=TradeContact(
        "Max Mustermann",
        "Sales",
        phone="+49 123 4567890",
        email="foo@example.com",
    ),
)

TEST_BUYER: Final = TradeParty(
    "Buyer AG",
    PostalAddress("DE"),
)

TEST_EN16931_INVOICE: Final = EN16931Invoice(
    "INV-12345",
    DocumentTypeCode.INVOICE,
    datetime.date(2024, 8, 20),
    TEST_SELLER,
    TEST_BUYER,
    "EUR",
    tax_basis_total_amount=Money("10090.00", "EUR"),
    tax_total_amounts=[Money("1917.10", "EUR")],
    grand_total_amount=Money("12007.10", "EUR"),
    due_payable_amount=Money("12007.10", "EUR"),
    line_total_amount=Money("10090.00", "EUR"),
    tax=[
        Tax(
            Money("1917.10", "EUR"),
            Money("10090.00", "EUR"),
            Decimal(19),
            TaxCategoryCode.STANDARD_RATE,
        ),
    ],
    delivery_date=datetime.date(2024, 8, 21),
    notes=[
        IncludedNote("This is a test invoice."),
        IncludedNote(
            "This is seller note.", TextSubjectCode.COMMENTS_BY_SELLER
        ),
    ],
    line_items=[
        LineItem(
            "1",
            "Fixed amount item\nWith multiple lines",
            Money("10000.00", "EUR"),
            (Decimal(1), QuantityCode.PIECE),
            Money("10000.00", "EUR"),
            Decimal("19"),
        ),
        EN16931LineItem(
            "2",
            "Hourly item",
            Money("30.00", "EUR"),
            (Decimal(3), QuantityCode.HOUR),
            Money("90.00", "EUR"),
            Decimal("19"),
            global_id=("9781529044195", "0160"),
            basis_quantity=(Decimal(1), QuantityCode.HOUR),
            description="This is a line item description.",
            note=IncludedNote("This is a line item note."),
            seller_assigned_id="ISBN-44",
            buyer_assigned_id="TT-123",
            product_characteristics=[
                ProductCharacteristic("color", "red"),
            ],
            product_classifications=[
                ProductClassification("1234-5679", list_id=ItemTypeCode.ISSN),
                ProductClassification(
                    "9781529044195",
                    list_id=ItemTypeCode.ISBN,
                    list_version_id="99",
                ),
            ],
            origin_country="DE",
            buyer_order_line_id="BUY-DOC",
            gross_unit_price=(
                Money("40.00", "EUR"),
                (Decimal(1), QuantityCode.HOUR),
            ),
            gross_allowance_or_charge=LineAllowance(
                Money("10.00", "EUR"),
                reason_code=AllowanceChargeCode.AHEAD_OF_SCHEDULE,
            ),
            charges=[
                LineCharge(
                    Money("0.05", "EUR"), reason="Complexity surcharge"
                ),
            ],
            allowances=[
                LineAllowance(
                    Money("1.00", "EUR"),
                    reason_code=AllowanceChargeCode.AHEAD_OF_SCHEDULE,
                    reason="Ahead of schedule",
                    basis_amount=Money("20.00", "EUR"),
                    percent=Decimal("5"),
                ),
            ],
            billing_period=(
                datetime.date(2024, 8, 1),
                datetime.date(2024, 8, 31),
            ),
            doc_ref=("REFDOC-1", None),
        ),
    ],
    buyer_reference="BUYER-1234",
    seller_order_id="SELL-DOC",
    buyer_order_id="BUY-DOC",
    contract_id="CONTRACT-123",
    referenced_docs=[
        ReferenceDocument("REFDOC-1", DocumentTypeCode.INVOICING_DATA_SHEET),
        ReferenceDocument(
            "REFDOC-2",
            DocumentTypeCode.RELATED_DOCUMENT,
            "Test ref doc",
            "https://example.com/refdoc.txt",
            attachment=(TEST_PNG, "image/png", "refdoc.png"),
            reference_type_code=ReferenceQualifierCode.PRICE_LIST_VERSION,
        ),
    ],
    procuring_project=("PROJ-123", "Project X"),
    sepa_reference="ABC-dddd",
    payment_means=[
        PaymentMeans(PaymentMeansCode.BANK_PAYMENT),
    ],
    payment_terms=PaymentTerms(
        due_date=datetime.date(2024, 9, 3),
    ),
)



"""
XML Utilities for PyCheval
"""

URN_MINIMUM_PROFILE = "urn:factur-x.eu:1p0:minimum"
URN_BASIC_WL_PROFILE = "urn:factur-x.eu:1p0:basicwl"
URN_BASIC_PROFILE = (
    "urn:cen.eu:en16931:2017#compliant#urn:factur-x.eu:1p0:basic"
)
URN_EN16931_PROFILE = "urn:cen.eu:en16931:2017"
URN_EXTENDED_PROFILE = (
    "urn:cen.eu:en16931:2017#conformant#urn:factur-x.eu:1p0:extended"
)
URN_XRECHNUNG_PROFILE = "urn:cen.eu:en16931:2017#compliant#urn:xoev-de:kosit:standard:xrechnung_2.1"  # noqa: E501

NS_CII = "urn:un:unece:uncefact:data:standard:CrossIndustryInvoice:100"
NS_RAM = "urn:un:unece:uncefact:data:standard:ReusableAggregateBusinessInformationEntity:100"  # noqa: E501
NS_UDT = "urn:un:unece:uncefact:data:standard:UnqualifiedDataType:100"

ALLOWED_ATTACHMENT_MIME_TYPES = (
    "application/pdf",
    "image/png",
    "image/jpeg",
    "text/csv",
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    "application/vnd.oasis.opendocument.spreadsheet",
)



from __future__ import annotations

import re
import xml.etree.ElementTree as ET
from base64 import b64decode
from collections.abc import Sequence
from datetime import date
from decimal import Decimal
from os import PathLike
from typing import TYPE_CHECKING, Any, NamedTuple

from .const import (
    NS_CII,
    NS_RAM,
    NS_UDT,
    URN_BASIC_PROFILE,
    URN_BASIC_WL_PROFILE,
    URN_EN16931_PROFILE,
    URN_EXTENDED_PROFILE,
    URN_MINIMUM_PROFILE,
    URN_XRECHNUNG_PROFILE,
)
from .exc import (
    InvalidProfileError,
    InvalidXMLError,
    NotFacturXError,
    UnsupportedProfileError,
    XMLParseError,
)
from .model import (
    BankAccount,
    BasicInvoice,
    BasicWLInvoice,
    DocumentAllowance,
    DocumentCharge,
    EN16931Invoice,
    EN16931LineItem,
    IncludedNote,
    LineAllowance,
    LineCharge,
    LineItem,
    MinimumInvoice,
    PaymentMeans,
    PaymentTerms,
    PostalAddress,
    ProductCharacteristic,
    ProductClassification,
    ReferenceDocument,
    Tax,
    TradeContact,
    TradeParty,
)
from .money import Money
from .quantities import QuantityCode
from .type_codes import (
    ALLOWED_MIME_TYPES,
    AllowanceChargeCode,
    DocumentTypeCode,
    IdentifierSchemeCode,
    ItemTypeCode,
    PaymentMeansCode,
    PaymentTimeCode,
    ReferenceQualifierCode,
    SpecialServiceCode,
    TaxCategoryCode,
    TextSubjectCode,
    VATExemptionCode,
)
from .types import ID, Attachment, DocRef, OptionalQuantity, Quantity

if TYPE_CHECKING:
    from xml.etree.ElementTree import _FileRead

    from _typeshed import StrPath


def _find_text(parent: ET.Element, tag: str) -> str:
    """Find the text of a child element with the given tag.

    Raise InvalidXMLError if the element is not found or if the element has
    no text.
    """
    el = parent.find(tag)
    if el is None:
        raise InvalidXMLError(f"Element {tag} not found")
    if el.text is None:
        raise InvalidXMLError(f"Element {tag} has no text")
    return el.text


def _find_all_texts(parent: ET.Element, tag: str) -> list[str]:
    texts = []
    for el in parent.findall(tag):
        if el.text is None:
            raise InvalidXMLError(f"Element {tag} has no text")
        texts.append(el.text)
    return texts


def _find_text_optional(parent: ET.Element, tag: str) -> str | None:
    """Find the text of a child element with the given tag.

    Return None if the element is not found. Raise InvalidXMLError if the
    element has no text.
    """
    el = parent.find(tag)
    if el is None:
        return None
    if el.text is None:
        raise InvalidXMLError(f"Element {tag} has no text")
    return el.text


def _find_indicator(parent: ET.Element, tag: str) -> bool:
    el = parent.find(tag)
    if el is None:
        raise InvalidXMLError(f"Element {tag} not found")
    indicator = el.find(f"./{{{NS_RAM}}}Indicator")
    if indicator is None:
        raise InvalidXMLError("Indicator element not found")
    if indicator.text == "true":
        return True
    elif indicator.text == "false":
        return False
    else:
        raise InvalidXMLError(f"Invalid indicator: {indicator.text}")


def _find_all_ids(
    parent: ET.Element, tag: str, *, scheme_required: bool = False
) -> list[ID]:
    ids: list[ID] = []
    for el in parent.findall(tag):
        ids.append(_parse_id(el, scheme_required=scheme_required))
    return ids


def _find_id_optional(
    parent: ET.Element, tag: str, *, scheme_required: bool = False
) -> ID | None:
    el = parent.find(tag)
    if el is None:
        return None
    return _parse_id(el, scheme_required=scheme_required)


def _parse_id(el: ET.Element, *, scheme_required: bool) -> ID:
    if el.text is None:
        raise InvalidXMLError("ID element has no text")
    scheme_id_s = el.attrib.get("schemeID")
    scheme_id: IdentifierSchemeCode | None = None
    if scheme_id_s is None and scheme_required:
        raise InvalidXMLError("schemeID attribute not found")
    elif scheme_id_s is not None:
        try:
            scheme_id = IdentifierSchemeCode(scheme_id_s)
        except ValueError as exc:
            raise InvalidXMLError(str(exc)) from exc
    return el.text, scheme_id


def _find_percent_optional(parent: ET.Element, tag: str) -> Decimal | None:
    rate_s = _find_text_optional(parent, tag)
    if rate_s is None:
        return None
    try:
        return Decimal(rate_s)
    except ValueError as exc:
        raise InvalidXMLError(f"Invalid tax rate: {rate_s}") from exc


def _find_quantity(parent: ET.Element, tag: str) -> Quantity:
    quantity = _find_optional_quantity_optional(parent, tag)
    if quantity is None:
        raise InvalidXMLError(f"Element {tag} not found")
    if quantity[1] is None:
        raise InvalidXMLError(f"Element {tag} has no unitCode")
    return quantity[0], quantity[1]


def _find_optional_quantity_optional(
    parent: ET.Element, tag: str
) -> OptionalQuantity | None:
    el = parent.find(tag)
    if el is None:
        return None
    if el.text is None:
        raise InvalidXMLError(f"Element {tag} has no text")
    try:
        quantity = Decimal(el.text)
    except ValueError as exc:
        raise InvalidXMLError(str(exc)) from exc
    code_s = el.attrib.get("unitCode")
    try:
        code = QuantityCode(code_s) if code_s is not None else None
    except ValueError as exc:
        raise InvalidXMLError(str(exc)) from exc
    return quantity, code


def _find_amount(
    parent: ET.Element,
    tag: str,
    default_currency: str,
    *,
    require_currency: bool = False,
) -> Money:
    el = parent.find(tag)
    if el is None:
        raise InvalidXMLError(f"Element {tag} not found")
    return _parse_amount(
        el, default_currency, require_currency=require_currency
    )


def _find_all_amounts(
    parent: ET.Element,
    tag: str,
    default_currency: str,
    *,
    require_currency: bool = False,
) -> list[Money]:
    return [
        _parse_amount(el, default_currency, require_currency=require_currency)
        for el in parent.findall(tag)
    ]


def _find_amount_optional(
    parent: ET.Element,
    tag: str,
    default_currency: str,
    *,
    require_currency: bool = False,
) -> Money | None:
    el = parent.find(tag)
    if el is None:
        return None
    return _parse_amount(
        el, default_currency, require_currency=require_currency
    )


def _parse_amount(
    el: ET.Element, default_currency: str, *, require_currency: bool = False
) -> Money:
    currency_code = el.attrib.get("currencyID")
    if currency_code is None and require_currency:
        raise InvalidXMLError("currencyID attribute not found")
    if currency_code is None:
        currency_code = default_currency
    amount = el.text
    if amount is None:
        raise InvalidXMLError("Element has no text")
    try:
        return Money(amount, currency_code)
    except ValueError as exc:
        raise InvalidXMLError(str(exc)) from exc


_DATE_RE = re.compile(r"\d{4}\d{2}\d{2}")


def _find_date(parent: ET.Element, tag: str) -> date:
    """Parse the text of a child element with the given tag as a date.

    Raise InvalidXMLError if the element is not found or if the element has
    no valid date.
    """
    date = _find_date_optional(parent, tag)
    if date is None:
        raise InvalidXMLError(f"Element {tag} not found")
    return date


def _find_date_optional(parent: ET.Element, tag: str) -> date | None:
    """Parse the text of a child element with the given tag as a date.

    Return None if the element is not found. Raise InvalidXMLError if the
    element has no valid date.
    """
    date_el = parent.find(tag)
    if date_el is None:
        return None
    dts_el = date_el.find(f"./{{{NS_UDT}}}DateTimeString")
    if dts_el is None:
        raise InvalidXMLError(f"DateTimeString element not found in {tag}")
    if dts_el.attrib.get("format") != "102":
        raise InvalidXMLError(f"Invalid format in DateTimeString in {tag}")
    text = dts_el.text
    if text is None:
        raise InvalidXMLError(f"DateTimeString element has no text in {tag}")
    if not _DATE_RE.fullmatch(text):
        raise InvalidXMLError(f"Invalid date: {text}")
    try:
        return date.fromisoformat(text)
    except ValueError:
        raise InvalidXMLError(f"Invalid date: {text}") from None


def _find_ref_doc_id_optional(parent: ET.Element, tag: str) -> str | None:
    """Find the ID of a referenced document.

    Return None if the element is not found. Raise InvalidXMLError if the
    element has no text.
    """
    el = parent.find(tag)
    if el is None:
        return None
    id_el = el.find(f"./{{{NS_RAM}}}IssuerAssignedID")
    if id_el is None:
        raise InvalidXMLError(f"IssuerAssignedID element not found in {tag}")
    if id_el.text is None:
        raise InvalidXMLError(f"IssuerAssignedID element has no text in {tag}")
    return id_el.text


def parse_xml(xml: str | _FileRead | StrPath) -> MinimumInvoice:
    """Parse a Factur-X XML file and return a matching invoice.

    Raise a FacturXParseError if the XML file is not a valid Factur-X file
    or a ModelError if the invoice is invalid.
    """

    tree = _parse_tree(xml)
    if tree.tag != f"{{{NS_CII}}}CrossIndustryInvoice":
        raise NotFacturXError("Root element is not a Factur-X invoice")
    id_el = tree.find(
        f"./{{{NS_CII}}}ExchangedDocumentContext/{{{NS_RAM}}}GuidelineSpecifiedDocumentContextParameter/{{{NS_RAM}}}ID"
    )
    if id_el is None:
        raise NotFacturXError("Profile ID element not found")
    if id_el.text == URN_MINIMUM_PROFILE:
        return _parse_minimum_invoice(tree)
    elif id_el.text == URN_BASIC_WL_PROFILE:
        return _parse_basic_wl_invoice(tree)
    elif id_el.text == URN_BASIC_PROFILE:
        return _parse_basic_invoice(tree)
    elif id_el.text == URN_EN16931_PROFILE:
        return _parse_en16931_invoice(tree)
    elif id_el.text == URN_EXTENDED_PROFILE:
        raise UnsupportedProfileError("Unsupported profile: EXTENDED")
    elif id_el.text == URN_XRECHNUNG_PROFILE:
        raise UnsupportedProfileError("Unsupported profile: XRECHNUNG")
    else:
        raise UnsupportedProfileError(f"Unsupported profile: {id_el.text}")


def _parse_tree(xml: str | _FileRead | StrPath) -> ET.Element:
    try:
        if isinstance(xml, str):
            return ET.fromstring(xml)
        elif isinstance(xml, (str, PathLike)):
            with open(xml) as f:
                return ET.parse(f).getroot()
        else:
            return ET.parse(xml).getroot()
    except ET.ParseError as exc:
        raise XMLParseError(str(exc)) from exc


def _parse_minimum_invoice(tree: ET.Element) -> MinimumInvoice:
    doc_ctx = _parse_doc_ctx(tree)
    doc_info = _parse_doc(tree)
    agreement, delivery, settlement = _parse_transaction(tree)
    if len(doc_info.notes) > 0:
        raise InvalidProfileError(
            "MINIMUM", "Included notes are not supported in MINIMUM profile"
        )
    if agreement.seller_tax_representative is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "SellerTaxRepresentativeTradeParty element is not supported in "
            "MINIMUM profile",
        )
    if agreement.contract_id is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "ContractReferencedDocument element is not supported in "
            "MINIMUM profile",
        )
    if agreement.seller_order_id is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "SellerOrderReferencedDocument element is not supported in "
            "MINIMUM profile",
        )
    if len(agreement.referenced_docs) > 0:
        raise InvalidProfileError(
            "MINIMUM",
            "AdditionalReferencedDocument element is not supported in "
            "MINIMUM profile",
        )
    if agreement.procuring_project is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "SpecifiedProcuringProject element is not supported in "
            "MINIMUM profile",
        )
    if delivery.ship_to is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "ShipToTradeParty element is not supported in MINIMUM profile",
        )
    if delivery.delivery_date is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "Delivery date is not supported in MINIMUM profile",
        )
    if delivery.despatch_advice_id is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "DespatchAdviceReferencedDocument element is not supported in "
            "MINIMUM profile",
        )
    if delivery.receiving_advice_id is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "ReceivingAdviceReferencedDocument element is not supported in "
            "MINIMUM profile",
        )
    if settlement.creditor_reference_id is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "CreditorReferenceID element is not supported in MINIMUM profile",
        )
    if settlement.payment_reference is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "PaymentReference element is not supported in MINIMUM profile",
        )
    if settlement.tax_currency_code is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "TaxCurrencyCode element is not supported in MINIMUM profile",
        )
    if settlement.payee is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "PayeeTradeParty element is not supported in MINIMUM profile",
        )
    if len(settlement.payment_means) > 0:
        raise InvalidProfileError(
            "MINIMUM",
            "SpecifiedTradeSettlementPaymentMeans element is not supported "
            "in MINIMUM profile",
        )
    if len(settlement.tax) > 0:
        raise InvalidProfileError(
            "MINIMUM",
            "ApplicableTradeTax element is not supported in MINIMUM profile",
        )
    if settlement.billing_period is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "Billing period is not supported in MINIMUM profile",
        )
    if len(settlement.allowances) > 0 or len(settlement.charges) > 0:
        raise InvalidProfileError(
            "MINIMUM",
            "SpecifiedTradeAllowanceCharge element is not supported in "
            "MINIMUM profile",
        )
    if settlement.payment_terms is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "SpecifiedTradePaymentTerms element is not supported in MINIMUM "
            "profile",
        )
    if settlement.summation.charge_total is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "ChargeTotalAmount element is not supported in MINIMUM profile",
        )
    if settlement.summation.allowance_total is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "AllowanceTotalAmount element is not supported in MINIMUM profile",
        )
    if settlement.summation.prepaid_amount is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "PrepaidAmount element is not supported in MINIMUM profile",
        )
    if settlement.summation.rounding_amount is not None:
        raise InvalidProfileError(
            "MINIMUM",
            "RoundingAmount element is not supported in MINIMUM profile",
        )
    if len(settlement.preceding_invoices) > 0:
        raise InvalidProfileError(
            "MINIMUM",
            "InvoiceReferencedDocument element is not supported in MINIMUM "
            "profile",
        )
    if len(settlement.receiver_accounting_ids) > 0:
        raise InvalidProfileError(
            "MINIMUM",
            "ReceivableSpecifiedTradeAccountingAccount element is not "
            "supported in MINIMUM profile",
        )
    return MinimumInvoice(
        **_minimum_args(doc_ctx, doc_info, agreement, settlement)
    )


def _parse_basic_wl_invoice(tree: ET.Element) -> BasicWLInvoice:
    doc_ctx = _parse_doc_ctx(tree)
    doc_info = _parse_doc(tree)
    agreement, delivery, settlement = _parse_transaction(tree)
    if agreement.seller_order_id is not None:
        raise InvalidProfileError(
            "BASIC WL",
            "SellerOrderReferencedDocument element is not supported in "
            "BASIC WL profile",
        )
    if len(agreement.referenced_docs) > 0:
        raise InvalidProfileError(
            "BASIC WL",
            "AdditionalReferencedDocument element is not supported in "
            "BASIC WL profile",
        )
    if agreement.procuring_project is not None:
        raise InvalidProfileError(
            "BASIC WL",
            "SpecifiedProcuringProject element is not supported in "
            "BASIC WL profile",
        )
    if delivery.receiving_advice_id is not None:
        raise InvalidProfileError(
            "BASIC WL",
            "ReceivingAdviceReferencedDocument element is not supported in "
            "BASIC WL profile",
        )
    if settlement.tax_currency_code is not None:
        raise InvalidProfileError(
            "BASIC WL",
            "TaxCurrencyCode element is not supported in BASIC WL profile",
        )
    if settlement.summation.rounding_amount is not None:
        raise InvalidProfileError(
            "BASIC WL",
            "RoundingAmount element is not supported in BASIC WL profile",
        )
    return BasicWLInvoice(
        **_minimum_args(doc_ctx, doc_info, agreement, settlement),
        **_basic_wl_args(doc_info, agreement, delivery, settlement),
    )


def _parse_basic_invoice(tree: ET.Element) -> BasicInvoice:
    doc_ctx = _parse_doc_ctx(tree)
    doc_info = _parse_doc(tree)
    agreement, delivery, settlement = _parse_transaction(tree)
    line_items = _parse_line_items(tree, settlement.currency_code)
    if agreement.seller_order_id is not None:
        raise InvalidProfileError(
            "BASIC",
            "SellerOrderReferencedDocument element is not supported in "
            "BASIC profile",
        )
    if len(agreement.referenced_docs) > 0:
        raise InvalidProfileError(
            "BASIC",
            "AdditionalReferencedDocument element is not supported in "
            "BASIC profile",
        )
    if agreement.procuring_project is not None:
        raise InvalidProfileError(
            "BASIC",
            "SpecifiedProcuringProject element is not supported in BASIC "
            "profile",
        )
    if delivery.receiving_advice_id is not None:
        raise InvalidProfileError(
            "BASIC",
            "ReceivingAdviceReferencedDocument element is not supported in "
            "BASIC profile",
        )
    if settlement.tax_currency_code is not None:
        raise InvalidProfileError(
            "BASIC",
            "TaxCurrencyCode element is not supported in BASIC profile",
        )
    if settlement.summation.rounding_amount is not None:
        raise InvalidProfileError(
            "BASIC",
            "RoundingAmount element is not supported in BASIC profile",
        )
    return BasicInvoice(
        **_minimum_args(doc_ctx, doc_info, agreement, settlement),
        **_basic_wl_args(doc_info, agreement, delivery, settlement),
        **_basic_args(line_items),
    )


def _parse_en16931_invoice(tree: ET.Element) -> EN16931Invoice:
    doc_ctx = _parse_doc_ctx(tree)
    doc_info = _parse_doc(tree)
    agreement, delivery, settlement = _parse_transaction(tree)
    line_items = _parse_en16931_line_items(tree, settlement.currency_code)
    return EN16931Invoice(
        **_minimum_args(doc_ctx, doc_info, agreement, settlement),
        **_basic_wl_args(doc_info, agreement, delivery, settlement),
        **_basic_args(line_items),
        **_en16931_args(agreement, delivery, settlement),
    )


def _minimum_args(
    doc_ctx: _DocumentContext,
    doc_info: _DocumentInfo,
    agreement: _TradeAgreement,
    settlement: _TradeSettlement,
) -> dict[str, Any]:
    return {
        "invoice_number": doc_info.id,
        "type_code": doc_info.type_code,
        "invoice_date": doc_info.issue_date,
        "seller": agreement.seller,
        "buyer": agreement.buyer,
        "currency_code": settlement.currency_code,
        "line_total_amount": settlement.summation.line_total,
        "tax_basis_total_amount": settlement.summation.tax_basis_total,
        "tax_total_amounts": settlement.summation.tax_totals,
        "grand_total_amount": settlement.summation.grand_total,
        "due_payable_amount": settlement.summation.due_payable_amount,
        "business_process_id": doc_ctx.business_process_id,
        "buyer_reference": agreement.buyer_reference,
        "buyer_order_id": agreement.buyer_order_id,
    }


def _basic_wl_args(
    doc_info: _DocumentInfo,
    agreement: _TradeAgreement,
    delivery: _TradeDelivery,
    settlement: _TradeSettlement,
) -> dict[str, Any]:
    return {
        "tax": settlement.tax,
        "charge_total_amount": settlement.summation.charge_total,
        "allowance_total_amount": settlement.summation.allowance_total,
        "prepaid_amount": settlement.summation.prepaid_amount,
        "payee": settlement.payee,
        "delivery_date": delivery.delivery_date,
        "billing_period": settlement.billing_period,
        "allowances": settlement.allowances,
        "charges": settlement.charges,
        "notes": doc_info.notes,
        "seller_tax_representative": agreement.seller_tax_representative,
        "contract_id": agreement.contract_id,
        "ship_to": delivery.ship_to,
        "despatch_advice_id": delivery.despatch_advice_id,
        "sepa_reference": settlement.creditor_reference_id,
        "payment_reference": settlement.payment_reference,
        "payment_means": settlement.payment_means,
        "payment_terms": settlement.payment_terms,
        "preceding_invoices": settlement.preceding_invoices,
        "receiver_accounting_ids": settlement.receiver_accounting_ids,
    }


def _basic_args(line_items: Sequence[LineItem]) -> dict[str, Any]:
    return {"line_items": line_items}


def _en16931_args(
    agreement: _TradeAgreement,
    delivery: _TradeDelivery,
    settlement: _TradeSettlement,
) -> dict[str, Any]:
    return {
        "receiving_advice_id": delivery.receiving_advice_id,
        "rounding_amount": settlement.summation.rounding_amount,
        "seller_order_id": agreement.seller_order_id,
        "procuring_project": agreement.procuring_project,
        "tax_currency_code": settlement.tax_currency_code,
        "referenced_docs": agreement.referenced_docs,
    }


class _DocumentContext(NamedTuple):
    business_process_id: str | None


def _parse_doc_ctx(tree: ET.Element) -> _DocumentContext:
    doc_ctx = tree.find(f"./{{{NS_CII}}}ExchangedDocumentContext")
    if doc_ctx is None:
        raise InvalidXMLError("ExchangedDocumentContext element not found")
    business_process_el = doc_ctx.find(
        f"./{{{NS_RAM}}}BusinessProcessSpecifiedDocumentContextParameter/{{{NS_RAM}}}ID"
    )
    process_id = (
        business_process_el.text if business_process_el is not None else None
    )
    return _DocumentContext(process_id)


class _DocumentInfo(NamedTuple):
    id: str
    type_code: DocumentTypeCode
    issue_date: date
    notes: list[IncludedNote]  # BASIC WL+


def _parse_doc(tree: ET.Element) -> _DocumentInfo:
    doc_el = tree.find(f"./{{{NS_CII}}}ExchangedDocument")
    if doc_el is None:
        raise InvalidXMLError("ExchangedDocument element not found")
    id = _find_text(doc_el, f"./{{{NS_RAM}}}ID")
    type_code_s = _find_text(doc_el, f"./{{{NS_RAM}}}TypeCode")
    try:
        type_code = DocumentTypeCode(int(type_code_s))
    except ValueError:
        raise InvalidXMLError(f"Invalid TypeCode: {type_code_s}") from None
    dt = _find_date(doc_el, f"./{{{NS_RAM}}}IssueDateTime")
    notes = [
        _parse_note(el) for el in doc_el.findall(f"./{{{NS_RAM}}}IncludedNote")
    ]
    return _DocumentInfo(id, type_code, dt, notes)


def _parse_note(el: ET.Element) -> IncludedNote:
    content = _find_text(el, f"./{{{NS_RAM}}}Content")
    subject_code: TextSubjectCode | None = None
    code_el = el.find(f"./{{{NS_RAM}}}SubjectCode")
    if code_el is not None:
        if code_el.text is None:
            raise InvalidXMLError("SubjectCode element has no text")
        try:
            subject_code = TextSubjectCode(code_el.text)
        except ValueError as exc:
            raise InvalidXMLError(
                f"Invalid SubjectCode: {code_el.text}"
            ) from exc
    return IncludedNote(content, subject_code)


def _parse_transaction(
    tree: ET.Element,
) -> tuple[_TradeAgreement, _TradeDelivery, _TradeSettlement]:
    el = tree.find(f"./{{{NS_CII}}}SupplyChainTradeTransaction")
    if el is None:
        raise InvalidXMLError("SupplyChainTradeTransaction element not found")
    agreement = _parse_agreement(el)
    delivery = _parse_delivery(el)
    settlement = _parse_settlement(el)
    return agreement, delivery, settlement


class _TradeAgreement(NamedTuple):
    seller: TradeParty
    buyer: TradeParty
    buyer_reference: str | None
    seller_order_id: str | None  # EN16931+
    buyer_order_id: str | None
    contract_id: str | None  # BASIC WL+
    seller_tax_representative: TradeParty | None  # BASIC WL+
    referenced_docs: Sequence[ReferenceDocument]  # EN16931+
    procuring_project: tuple[str, str] | None  # EN16931+


def _parse_agreement(parent: ET.Element) -> _TradeAgreement:
    el = parent.find(f"./{{{NS_RAM}}}ApplicableHeaderTradeAgreement")
    if el is None:
        raise InvalidXMLError(
            "ApplicableHeaderTradeAgreement etax_currency_codelement not found"
        )
    buyer_ref = _find_text_optional(el, f"./{{{NS_RAM}}}BuyerReference")
    seller = _parse_trade_party(el, f"./{{{NS_RAM}}}SellerTradeParty")
    buyer = _parse_trade_party(el, f"./{{{NS_RAM}}}BuyerTradeParty")
    seller_order_id = _find_ref_doc_id_optional(
        el, f"./{{{NS_RAM}}}SellerOrderReferencedDocument"
    )
    buyer_order_id = _find_ref_doc_id_optional(
        el, f"./{{{NS_RAM}}}BuyerOrderReferencedDocument"
    )
    seller_tax_representative = _parse_trade_party_optional(
        el, f"./{{{NS_RAM}}}SellerTaxRepresentativeTradeParty"
    )
    contract_id = _find_ref_doc_id_optional(
        el, f"./{{{NS_RAM}}}ContractReferencedDocument"
    )
    referenced_docs = [
        _parse_reference_document(doc_el)
        for doc_el in el.findall(f"./{{{NS_RAM}}}AdditionalReferencedDocument")
    ]
    project: tuple[str, str] | None = None
    project_el = el.find(f"./{{{NS_RAM}}}SpecifiedProcuringProject")
    if project_el is not None:
        project = (
            _find_text(project_el, f"./{{{NS_RAM}}}ID"),
            _find_text(project_el, f"./{{{NS_RAM}}}Name"),
        )
    return _TradeAgreement(
        seller,
        buyer,
        buyer_ref,
        seller_order_id,
        buyer_order_id,
        contract_id,
        seller_tax_representative,
        referenced_docs,
        project,
    )


class _TradeDelivery(NamedTuple):
    ship_to: TradeParty | None  # BASIC WL+
    delivery_date: date | None  # BASIC WL+
    despatch_advice_id: str | None  # BASIC WL+
    receiving_advice_id: str | None  # EN16931+


def _parse_delivery(parent: ET.Element) -> _TradeDelivery:
    el = parent.find(f"./{{{NS_RAM}}}ApplicableHeaderTradeDelivery")
    if el is None:
        raise InvalidXMLError(
            "ApplicableHeaderTradeDelivery element not found"
        )
    ship_to = _parse_trade_party_optional(
        el, f"./{{{NS_RAM}}}ShipToTradeParty"
    )
    delivery_date = _parse_delivery_date(el)
    despatch_advice_id = _find_ref_doc_id_optional(
        el, f"./{{{NS_RAM}}}DespatchAdviceReferencedDocument"
    )
    receiving_advice_id = _find_ref_doc_id_optional(
        el, f"./{{{NS_RAM}}}ReceivingAdviceReferencedDocument"
    )
    return _TradeDelivery(
        ship_to, delivery_date, despatch_advice_id, receiving_advice_id
    )


def _parse_delivery_date(parent: ET.Element) -> date | None:
    el = parent.find(f"./{{{NS_RAM}}}ActualDeliverySupplyChainEvent")
    if el is None:
        return None
    return _find_date(el, f"./{{{NS_RAM}}}OccurrenceDateTime")


class _SettlementSummation(NamedTuple):
    line_total: Money | None
    tax_basis_total: Money
    tax_totals: list[Money]
    grand_total: Money
    due_payable_amount: Money
    charge_total: Money | None  # BASIC WL+
    allowance_total: Money | None  # BASIC WL+
    prepaid_amount: Money | None  # BASIC WL+
    rounding_amount: Money | None  # EN16931+


class _TradeSettlement(NamedTuple):
    currency_code: str
    summation: _SettlementSummation
    creditor_reference_id: str | None  # BASIC WL+
    payment_reference: str | None  # BASIC WL+
    tax_currency_code: str | None  # EN16931+
    payee: TradeParty | None  # BASIC WL+
    payment_means: list[PaymentMeans]  # BASIC WL+
    tax: list[Tax]  # BASIC WL+
    billing_period: tuple[date, date] | None  # BASIC WL+
    allowances: list[DocumentAllowance]  # BASIC WL+
    charges: list[DocumentCharge]  # BASIC WL+
    payment_terms: PaymentTerms | None  # BASIC WL+
    preceding_invoices: list[tuple[str, date | None]]  # BASIC WL+
    receiver_accounting_ids: list[str]  # BASIC WL+


def _parse_settlement(parent: ET.Element) -> _TradeSettlement:
    el = parent.find(f"./{{{NS_RAM}}}ApplicableHeaderTradeSettlement")
    if el is None:
        raise InvalidXMLError(
            "ApplicableHeaderTradeSettlement element not found"
        )
    currency_code = _find_text(el, f"./{{{NS_RAM}}}InvoiceCurrencyCode")
    creditor_reference_id = _find_text_optional(
        el, f"./{{{NS_RAM}}}CreditorReferenceID"
    )
    payment_reference = _find_text_optional(
        el, f"./{{{NS_RAM}}}PaymentReference"
    )
    tax_currency_code = _find_text_optional(
        el, f"./{{{NS_RAM}}}TaxCurrencyCode"
    )
    payee = _parse_trade_party_optional(el, f"./{{{NS_RAM}}}PayeeTradeParty")
    payment_means = [
        _parse_payment_means(pay_el)
        for pay_el in el.findall(
            f"./{{{NS_RAM}}}SpecifiedTradeSettlementPaymentMeans"
        )
    ]
    tax = [
        _parse_tax(tax_el, currency_code)
        for tax_el in el.findall(f"./{{{NS_RAM}}}ApplicableTradeTax")
    ]
    billing_period = _parse_billing_period_optional(el)
    allowances = [
        _parse_allowance_or_charge(ac_el, currency_code)
        for ac_el in el.findall(f"./{{{NS_RAM}}}SpecifiedTradeAllowanceCharge")
    ]
    payment_terms = _parse_payment_terms(el)
    summation = _parse_summation(el, currency_code)
    referenced_invoices = _parse_referenced_invoices(el)
    receiver_accounting_ids: list[str] = []
    for el in parent.findall(
        f"./{{{NS_RAM}}}ReceivableSpecifiedTradeAccountingAccount"
    ):
        receiver_accounting_ids.append(_find_text(el, f"./{{{NS_RAM}}}ID"))
    return _TradeSettlement(
        currency_code,
        summation,
        creditor_reference_id,
        payment_reference,
        tax_currency_code,
        payee,
        payment_means,
        tax,
        billing_period,
        [a for a in allowances if isinstance(a, DocumentAllowance)],
        [c for c in allowances if isinstance(c, DocumentCharge)],
        payment_terms,
        referenced_invoices,
        receiver_accounting_ids,
    )


def _parse_payment_means(el: ET.Element) -> PaymentMeans:
    type_code = _parse_payment_type_code(el)
    information = _find_text_optional(el, f"./{{{NS_RAM}}}Information")

    card: tuple[str, str | None] | None = None
    card_el = el.find(f"./{{{NS_RAM}}}ApplicableTradeSettlementFinancialCard")
    if card_el is not None:
        card = (
            _find_text(card_el, f"./{{{NS_RAM}}}ID"),
            _find_text_optional(card_el, f"./{{{NS_RAM}}}CardholderName"),
        )

    payer_iban: str | None = None
    payer_el = el.find(f"./{{{NS_RAM}}}PayerPartyDebtorFinancialAccount")
    if payer_el is not None:
        payer_iban_el = payer_el.find(f"./{{{NS_RAM}}}IBANID")
        if payer_iban_el is not None:
            payer_iban = payer_iban_el.text
    payee_account = _parse_account_optional(
        el, f"./{{{NS_RAM}}}PayeePartyCreditorFinancialAccount"
    )
    payee_bic = _find_text_optional(
        el,
        f"./{{{NS_RAM}}}PayeeSpecifiedCreditorFinancialInstitution/{{{NS_RAM}}}BICID",
    )
    return PaymentMeans(
        type_code,
        payee_account,
        payee_bic,
        information,
        card=card,
        payer_iban=payer_iban,
    )


def _parse_payment_type_code(parent: ET.Element) -> PaymentMeansCode:
    type_code_s = _find_text(parent, f"./{{{NS_RAM}}}TypeCode")
    try:
        return PaymentMeansCode(type_code_s)
    except ValueError as exc:
        raise InvalidXMLError(str(exc)) from exc


def _parse_account_optional(
    parent: ET.Element, tag: str
) -> BankAccount | None:
    el = parent.find(tag)
    if el is None:
        return None
    iban = _find_text_optional(el, f"./{{{NS_RAM}}}IBANID")
    account_name = _find_text_optional(el, f"./{{{NS_RAM}}}AccountName")
    bank_id = _find_text_optional(el, f"./{{{NS_RAM}}}ProprietaryID")
    return BankAccount(iban, account_name, bank_id)


def _parse_line_items(
    parent: ET.Element, default_currency: str
) -> list[LineItem]:
    el = parent.find(f"./{{{NS_CII}}}SupplyChainTradeTransaction")
    if el is None:
        raise InvalidXMLError("SupplyChainTradeTransaction element not found")
    return [
        _parse_line_item(li_el, default_currency)
        for li_el in el.findall(
            f"./{{{NS_RAM}}}IncludedSupplyChainTradeLineItem"
        )
    ]


def _parse_en16931_line_items(
    parent: ET.Element, default_currency: str
) -> list[EN16931LineItem]:
    el = parent.find(f"./{{{NS_CII}}}SupplyChainTradeTransaction")
    if el is None:
        raise InvalidXMLError("SupplyChainTradeTransaction element not found")
    return [
        _parse_en16931_line_item(li_el, default_currency)
        for li_el in el.findall(
            f"./{{{NS_RAM}}}IncludedSupplyChainTradeLineItem"
        )
    ]


def _parse_line_item(el: ET.Element, default_currency: str) -> LineItem:
    doc = _parse_line_document(el)
    product = _parse_trade_product(el)
    agreement = _parse_line_agreement(el, default_currency)
    delivery = _parse_line_delivery(el)
    settlement = _parse_line_settlement(el, default_currency)

    if doc.note is not None:
        raise InvalidProfileError(
            "BASIC",
            "Included notes are not supported in BASIC profile line items",
        )
    if product.seller_id is not None:
        raise InvalidProfileError(
            "BASIC",
            "SellerAssignedID element is not supported in BASIC profile "
            "line items",
        )
    if product.buyer_id is not None:
        raise InvalidProfileError(
            "BASIC",
            "BuyerAssignedID element is not supported in BASIC profile "
            "line items",
        )
    if product.description is not None:
        raise InvalidProfileError(
            "BASIC",
            "Description element is not supported in BASIC profile line items",
        )
    if len(product.characteristics) > 0:
        raise InvalidProfileError(
            "BASIC",
            "ApplicableProductCharacteristic element is not supported in "
            "BASIC profile line items",
        )
    if len(product.classifications) > 0:
        raise InvalidProfileError(
            "BASIC",
            "DesignatedProductClassification element is not supported in "
            "BASIC profile line items",
        )
    if product.origin_country is not None:
        raise InvalidProfileError(
            "BASIC",
            "OriginCountry element is not supported in BASIC profile "
            "line items",
        )
    if agreement.buyer_order_line_id is not None:
        raise InvalidProfileError(
            "BASIC",
            "BuyerOrderReferencedDocument element is not supported in BASIC "
            "profile line items",
        )
    if agreement.gross_unit_price is not None:
        raise InvalidProfileError(
            "BASIC",
            "GrossPriceProductTradePrice element is not supported in BASIC "
            "profile line items",
        )
    if settlement.billing_period is not None:
        raise InvalidProfileError(
            "BASIC",
            "BillingSpecifiedPeriod is not supported in BASIC profile "
            "line items",
        )
    if settlement.doc_ref is not None:
        raise InvalidProfileError(
            "BASIC",
            "AdditionalReferencedDocument element is not supported in "
            "BASIC profile line items",
        )
    if settlement.trade_account_id is not None:
        raise InvalidProfileError(
            "BASIC",
            "ReceivableSpecifiedTradeAccountingAccount element is not "
            "supported in BASIC profile line items",
        )

    return LineItem(
        doc.id,
        product.name,
        agreement.net_price,
        delivery.billed_quantity,
        settlement.total_amount,
        settlement.tax_rate,
        settlement.tax_category,
        global_id=product.global_id,
        basis_quantity=agreement.basis_quantity,
        allowances=settlement.allowances,
        charges=settlement.charges,
    )


def _parse_en16931_line_item(
    el: ET.Element, default_currency: str
) -> EN16931LineItem:
    doc = _parse_line_document(el)
    product = _parse_trade_product(el)
    agreement = _parse_line_agreement(el, default_currency)
    delivery = _parse_line_delivery(el)
    settlement = _parse_line_settlement(el, default_currency)
    return EN16931LineItem(
        doc.id,
        product.name,
        agreement.net_price,
        delivery.billed_quantity,
        settlement.total_amount,
        settlement.tax_rate,
        settlement.tax_category,
        global_id=product.global_id,
        basis_quantity=agreement.basis_quantity,
        gross_unit_price=agreement.gross_unit_price,
        allowances=settlement.allowances,
        charges=settlement.charges,
        note=doc.note,
        seller_assigned_id=product.seller_id,
        buyer_assigned_id=product.buyer_id,
        description=product.description,
        product_characteristics=product.characteristics,
        product_classifications=product.classifications,
        origin_country=product.origin_country,
        buyer_order_line_id=agreement.buyer_order_line_id,
        billing_period=settlement.billing_period,
        doc_ref=settlement.doc_ref,
        trade_account_id=settlement.trade_account_id,
    )


class _LineDocument(NamedTuple):
    id: str
    note: IncludedNote | None  # EN16931+


def _parse_line_document(parent: ET.Element) -> _LineDocument:
    el = parent.find(f"./{{{NS_RAM}}}AssociatedDocumentLineDocument")
    if el is None:
        raise InvalidXMLError(
            "AssociatedDocumentLineDocument element not found"
        )
    id = _find_text(el, f"./{{{NS_RAM}}}LineID")

    note: IncludedNote | None = None
    note_el = el.find(f"./{{{NS_RAM}}}IncludedNote")
    if note_el is not None:
        note = _parse_note(note_el)

    return _LineDocument(id, note)


class _TradeProduct(NamedTuple):
    name: str
    global_id: ID | None
    seller_id: str | None  # EN16931+
    buyer_id: str | None  # EN16931+
    description: str | None  # EN16931+
    characteristics: list[ProductCharacteristic]  # EN16931+
    classifications: list[ProductClassification]  # EN16931+
    origin_country: str | None  # EN16931+


def _parse_trade_product(parent: ET.Element) -> _TradeProduct:
    el = parent.find(f"./{{{NS_RAM}}}SpecifiedTradeProduct")
    if el is None:
        raise InvalidXMLError("SpecifiedTradeProduct element not found")
    global_id = _find_id_optional(el, f"./{{{NS_RAM}}}GlobalID")
    seller_id = _find_text_optional(el, f"./{{{NS_RAM}}}SellerAssignedID")
    buyer_id = _find_text_optional(el, f"./{{{NS_RAM}}}BuyerAssignedID")
    name = _find_text(el, f"./{{{NS_RAM}}}Name")
    description = _find_text_optional(el, f"./{{{NS_RAM}}}Description")
    characteristics = [
        _parse_product_characteristic(pc_el)
        for pc_el in el.findall(
            f"./{{{NS_RAM}}}ApplicableProductCharacteristic"
        )
    ]
    classifications = [
        _parse_product_classification(pc_el)
        for pc_el in el.findall(
            f"./{{{NS_RAM}}}DesignatedProductClassification"
        )
    ]
    origin_country = _find_text_optional(
        el, f"./{{{NS_RAM}}}OriginCountry/{{{NS_RAM}}}ID"
    )
    return _TradeProduct(
        name,
        global_id,
        seller_id,
        buyer_id,
        description,
        characteristics,
        classifications,
        origin_country,
    )


def _parse_product_characteristic(parent: ET.Element) -> ProductCharacteristic:
    description = _find_text(parent, f"./{{{NS_RAM}}}Description")
    value = _find_text(parent, f"./{{{NS_RAM}}}Value")
    return ProductCharacteristic(description, value)


def _parse_product_classification(parent: ET.Element) -> ProductClassification:
    class_code_el = parent.find(f"./{{{NS_RAM}}}ClassCode")
    if class_code_el is None:
        raise InvalidXMLError("ClassCode element not found")
    class_code = class_code_el.text
    if class_code is None:
        raise InvalidXMLError("ClassCode element has no text")
    if "listID" not in class_code_el.attrib:
        raise InvalidXMLError("ClassCode element has no listID attribute")
    list_id_s = class_code_el.attrib["listID"]
    list_version_id = class_code_el.attrib.get("listVersionID")

    try:
        list_id = ItemTypeCode(list_id_s)
    except ValueError as exc:
        raise InvalidXMLError(str(exc)) from exc

    return ProductClassification(
        class_code, list_id=list_id, list_version_id=list_version_id
    )


class _LineAgreement(NamedTuple):
    gross_unit_price: tuple[Money, OptionalQuantity | None] | None  # EN16931+
    gross_allowance_or_charge: LineAllowance | LineCharge | None  # EN16931+
    net_price: Money
    basis_quantity: OptionalQuantity | None
    buyer_order_line_id: str | None  # EN16931+


def _parse_line_agreement(
    parent: ET.Element, default_currency: str
) -> _LineAgreement:
    el = parent.find(f"./{{{NS_RAM}}}SpecifiedLineTradeAgreement")
    if el is None:
        raise InvalidXMLError("SpecifiedLineTradeAgreement element not found")

    (gross_price, gross_charge) = _parse_gross_line_price(el, default_currency)

    net_price_el = el.find(f"./{{{NS_RAM}}}NetPriceProductTradePrice")
    if net_price_el is None:
        raise InvalidXMLError("NetPriceProductTradePrice element not found")
    net_price = _find_amount(
        net_price_el, f"./{{{NS_RAM}}}ChargeAmount", default_currency
    )
    basis_quantity = _find_optional_quantity_optional(
        el, f"./{{{NS_RAM}}}BasisQuantity"
    )
    buyer_order_line_id = _find_text_optional(
        el, f"./{{{NS_RAM}}}BuyerOrderReferencedDocument/{{{NS_RAM}}}LineID"
    )
    return _LineAgreement(
        gross_price,
        gross_charge,
        net_price,
        basis_quantity,
        buyer_order_line_id,
    )


def _parse_gross_line_price(
    parent: ET.Element, default_currency: str
) -> tuple[
    tuple[Money, OptionalQuantity | None] | None,
    LineAllowance | LineCharge | None,
]:
    el = parent.find(f"./{{{NS_RAM}}}GrossPriceProductTradePrice")
    if el is None:
        return None, None
    gross_amount = _find_amount(
        el, f"./{{{NS_RAM}}}ChargeAmount", default_currency
    )
    gross_quantity = _find_optional_quantity_optional(
        el, f"./{{{NS_RAM}}}BasisQuantity"
    )
    allowance: LineAllowance | LineCharge | None = None
    allowance_el = el.find(f"./{{{NS_RAM}}}AppliedTradeAllowanceCharge")
    if allowance_el is not None:
        allowance = _parse_line_allowance_or_charge(
            allowance_el, default_currency
        )
        if allowance.reason is not None:
            raise InvalidProfileError(
                "EN16931",
                "Reason element is not supported in "
                "AppliedTradeAllowanceCharge",
            )
    return (gross_amount, gross_quantity), allowance


class _LineDelivery(NamedTuple):
    billed_quantity: Quantity


def _parse_line_delivery(parent: ET.Element) -> _LineDelivery:
    el = parent.find(f"./{{{NS_RAM}}}SpecifiedLineTradeDelivery")
    if el is None:
        raise InvalidXMLError("SpecifiedLineTradeDelivery element not found")
    billed_quantity = _find_quantity(el, f"./{{{NS_RAM}}}BilledQuantity")
    return _LineDelivery(billed_quantity)


class _LineSettlement(NamedTuple):
    tax_category: TaxCategoryCode
    tax_rate: Decimal | None
    allowances: Sequence[LineAllowance]
    charges: Sequence[LineCharge]
    total_amount: Money
    billing_period: tuple[date, date] | None  # EN16931+
    doc_ref: DocRef | None  # EN16931+
    trade_account_id: str | None  # EN16931+


def _parse_line_settlement(
    parent: ET.Element, default_currency: str
) -> _LineSettlement:
    el = parent.find(f"./{{{NS_RAM}}}SpecifiedLineTradeSettlement")
    if el is None:
        raise InvalidXMLError("SpecifiedLineTradeSettlement element not found")
    tax_el = el.find(f"./{{{NS_RAM}}}ApplicableTradeTax")
    if tax_el is None:
        raise InvalidXMLError("ApplicableTradeTax element not found")
    tax_category, tax_rate = _parse_tax_simple(tax_el)
    allowances = [
        _parse_line_allowance_or_charge(ac_el, default_currency)
        for ac_el in el.findall(f"./{{{NS_RAM}}}SpecifiedTradeAllowanceCharge")
    ]
    total_amount = _parse_line_summation(el, default_currency)
    billing_period = _parse_billing_period_optional(el)

    doc_ref: DocRef | None = None
    ref_doc_el = el.find(f"./{{{NS_RAM}}}AdditionalReferencedDocument")
    if ref_doc_el is not None:
        doc_ref = _parse_doc_ref(ref_doc_el)

    trade_account_id = _find_text_optional(
        el,
        f"./{{{NS_RAM}}}ReceivableSpecifiedTradeAccountingAccount/{{{NS_RAM}}}ID",
    )

    return _LineSettlement(
        tax_category,
        tax_rate,
        [a for a in allowances if isinstance(a, LineAllowance)],
        [c for c in allowances if isinstance(c, LineCharge)],
        total_amount,
        billing_period,
        doc_ref,
        trade_account_id,
    )


def _parse_line_summation(parent: ET.Element, default_currency: str) -> Money:
    el = parent.find(
        f"./{{{NS_RAM}}}SpecifiedTradeSettlementLineMonetarySummation"
    )
    if el is None:
        raise InvalidXMLError(
            "SpecifiedTradeSettlementLineMonetarySummation element not found"
        )
    return _find_amount(el, f"./{{{NS_RAM}}}LineTotalAmount", default_currency)


def _parse_line_allowance_or_charge(
    el: ET.Element, default_currency: str
) -> LineAllowance | LineCharge:
    surcharge = _find_indicator(el, f"./{{{NS_RAM}}}ChargeIndicator")
    percent = _find_percent_optional(el, f"./{{{NS_RAM}}}CalculationPercent")
    basis_amount = _find_amount_optional(
        el, f"./{{{NS_RAM}}}BasisAmount", default_currency
    )
    actual_amount = _find_amount(
        el, f"./{{{NS_RAM}}}ActualAmount", default_currency
    )
    service_code: SpecialServiceCode | None = None
    allowance_code: AllowanceChargeCode | None = None
    reason_code_s = _find_text_optional(el, f"./{{{NS_RAM}}}ReasonCode")
    if reason_code_s is not None:
        try:
            if surcharge:
                service_code = SpecialServiceCode(reason_code_s)
            else:
                allowance_code = AllowanceChargeCode(int(reason_code_s))
        except ValueError as exc:
            raise InvalidXMLError(str(exc)) from exc
    reason = _find_text_optional(el, f"./{{{NS_RAM}}}Reason")
    if surcharge:
        return LineCharge(
            actual_amount,
            service_code,
            reason,
            percent=percent,
            basis_amount=basis_amount,
        )
    else:
        return LineAllowance(
            actual_amount,
            allowance_code,
            reason,
            percent=percent,
            basis_amount=basis_amount,
        )


def _parse_allowance_or_charge(
    el: ET.Element, default_currency: str
) -> DocumentAllowance | DocumentCharge:
    line_allowance = _parse_line_allowance_or_charge(el, default_currency)
    tax_el = el.find(f"./{{{NS_RAM}}}CategoryTradeTax")
    if tax_el is None:
        raise InvalidXMLError("CategoryTradeTax element not found")
    tax_type_code = _find_text(tax_el, f"./{{{NS_RAM}}}TypeCode")
    if tax_type_code != "VAT":
        raise InvalidXMLError(f"Invalid tax TypeCode: {tax_type_code}")
    tax_category_s = _find_text(tax_el, f"./{{{NS_RAM}}}CategoryCode")
    try:
        tax_category = TaxCategoryCode(tax_category_s)
    except ValueError as exc:
        raise InvalidXMLError(str(exc)) from exc
    tax_rate = _find_percent_optional(
        tax_el, f"./{{{NS_RAM}}}RateApplicablePercent"
    )
    if isinstance(line_allowance, LineCharge):
        return DocumentCharge(
            line_allowance.actual_amount,
            line_allowance.reason_code,
            line_allowance.reason,
            percent=line_allowance.percent,
            basis_amount=line_allowance.basis_amount,
            tax_category=tax_category,
            tax_rate=tax_rate,
        )
    else:
        return DocumentAllowance(
            line_allowance.actual_amount,
            line_allowance.reason_code,
            line_allowance.reason,
            percent=line_allowance.percent,
            basis_amount=line_allowance.basis_amount,
            tax_category=tax_category,
            tax_rate=tax_rate,
        )


def _parse_payment_terms(parent: ET.Element) -> PaymentTerms | None:
    el = parent.find(f"./{{{NS_RAM}}}SpecifiedTradePaymentTerms")
    if el is None:
        return None
    description = _find_text_optional(el, f"./{{{NS_RAM}}}Description")
    due_date = _find_date_optional(el, f"./{{{NS_RAM}}}DueDateDateTime")
    mandate_id = _find_text_optional(el, f"./{{{NS_RAM}}}DirectDebitMandateID")
    return PaymentTerms(
        description=description,
        due_date=due_date,
        direct_debit_mandate_id=mandate_id,
    )


def _parse_summation(
    parent: ET.Element, default_currency: str
) -> _SettlementSummation:
    el = parent.find(
        f"./{{{NS_RAM}}}SpecifiedTradeSettlementHeaderMonetarySummation"
    )
    if el is None:
        raise InvalidXMLError(
            "SpecifiedTradeSettlementHeaderMonetarySummation element not found"
        )
    line_total = _find_amount_optional(
        el, f"./{{{NS_RAM}}}LineTotalAmount", default_currency
    )
    charge_total = _find_amount_optional(
        el, f"./{{{NS_RAM}}}ChargeTotalAmount", default_currency
    )
    allowance_total = _find_amount_optional(
        el, f"./{{{NS_RAM}}}AllowanceTotalAmount", default_currency
    )
    tax_basis_total = _find_amount(
        el, f"./{{{NS_RAM}}}TaxBasisTotalAmount", default_currency
    )
    tax_totals = _find_all_amounts(
        el, f"./{{{NS_RAM}}}TaxTotalAmount", default_currency
    )
    rounding_amount = _find_amount_optional(
        el, f"./{{{NS_RAM}}}RoundingAmount", default_currency
    )
    grand_total = _find_amount(
        el, f"./{{{NS_RAM}}}GrandTotalAmount", default_currency
    )
    prepaid = _find_amount_optional(
        el, f"./{{{NS_RAM}}}TotalPrepaidAmount", default_currency
    )
    due_payable = _find_amount(
        el, f"./{{{NS_RAM}}}DuePayableAmount", default_currency
    )
    return _SettlementSummation(
        line_total,
        tax_basis_total,
        tax_totals,
        grand_total,
        due_payable,
        charge_total,
        allowance_total,
        prepaid,
        rounding_amount,
    )


def _parse_referenced_invoices(
    parent: ET.Element,
) -> list[tuple[str, date | None]]:
    els = parent.findall(f"./{{{NS_RAM}}}InvoiceReferencedDocument")
    return [_parse_referenced_invoice(el) for el in els]


def _parse_referenced_invoice(el: ET.Element) -> tuple[str, date | None]:
    id = _find_text(el, f"./{{{NS_RAM}}}IssuerAssignedID")
    issue_date = _find_date_optional(
        el, f"./{{{NS_RAM}}}FormattedIssueDateTime"
    )
    return id, issue_date


# Parsing recurring elements


def _parse_trade_party(parent: ET.Element, tag: str) -> TradeParty:
    trade_party = _parse_trade_party_optional(parent, tag)
    if trade_party is None:
        raise InvalidXMLError(f"{tag} element not found")
    return trade_party


def _parse_trade_party_optional(
    parent: ET.Element, tag: str
) -> TradeParty | None:
    party_el = parent.find(tag)
    if party_el is None:
        return None
    ids = _find_all_texts(party_el, f"./{{{NS_RAM}}}ID")
    global_ids = _find_all_ids(party_el, f"./{{{NS_RAM}}}GlobalID")
    name = _find_text(party_el, f"./{{{NS_RAM}}}Name")
    description = _find_text_optional(party_el, f"./{{{NS_RAM}}}Description")
    legal_org_id, legal_org_name = _parse_legal_org(party_el)
    contact = _parse_trade_contact(party_el)
    address = _parse_address(party_el)
    email = _parse_email(party_el)
    tax_number: str | None = None
    vat_id: str | None = None
    for tax_reg in party_el.findall(f"./{{{NS_RAM}}}SpecifiedTaxRegistration"):
        is_vat_id, tax_reg_id = _parse_tax_reg(tax_reg)
        if is_vat_id:
            if vat_id is not None:
                raise InvalidXMLError("Multiple VAT IDs found")
            vat_id = tax_reg_id
        else:
            if tax_number is not None:
                raise InvalidXMLError("Multiple tax numbers found")
            tax_number = tax_reg_id

    return TradeParty(
        name,
        address,
        email,
        description=description,
        tax_number=tax_number,
        vat_id=vat_id,
        ids=ids,
        global_ids=global_ids,
        legal_id=legal_org_id,
        trading_business_name=legal_org_name,
        contact=contact,
    )


def _parse_legal_org(parent: ET.Element) -> tuple[ID | None, str | None]:
    legal_org_el = parent.find(f"./{{{NS_RAM}}}SpecifiedLegalOrganization")
    if legal_org_el is None:
        return None, None
    id = _find_id_optional(legal_org_el, f"./{{{NS_RAM}}}ID")
    name = _find_text_optional(
        legal_org_el, f"./{{{NS_RAM}}}TradingBusinessName"
    )
    return id, name


def _parse_tax_reg(parent: ET.Element) -> tuple[bool, str]:
    """Parse a SpecifiedTaxRegistration element.

    Return a tuple with a boolean indicating whether the ID is a VAT ID (True)
    or a tax number (False) and the ID itself.
    """
    id_el = parent.find(f"./{{{NS_RAM}}}ID")
    if id_el is None:
        raise InvalidXMLError(
            "ID element not found in SpecifiedTaxRegistration"
        )
    scheme = id_el.attrib.get("schemeID")
    id = id_el.text
    if id is None:
        raise InvalidXMLError("ID element has no text")
    match scheme:
        case "FC":
            return False, id
        case "VA":
            return True, id
        case _:
            raise InvalidXMLError(f"Invalid schemeID: {scheme}")


def _parse_trade_contact(parent: ET.Element) -> TradeContact | None:
    el = parent.find(f"./{{{NS_RAM}}}DefinedTradeContact")
    if el is None:
        return None

    person_name = _find_text_optional(el, f"./{{{NS_RAM}}}PersonName")
    department_name = _find_text_optional(el, f"./{{{NS_RAM}}}DepartmentName")
    phone = _find_text_optional(
        el,
        f"./{{{NS_RAM}}}TelephoneUniversalCommunication/{{{NS_RAM}}}CompleteNumber",
    )
    email = _find_text_optional(
        el, f"./{{{NS_RAM}}}EmailURIUniversalCommunication/{{{NS_RAM}}}URIID"
    )
    return TradeContact(
        person_name=person_name,
        department_name=department_name,
        phone=phone,
        email=email,
    )


def _parse_address(parent: ET.Element) -> PostalAddress | None:
    address_el = parent.find(f"./{{{NS_RAM}}}PostalTradeAddress")
    if address_el is None:
        return None
    post_code = _find_text_optional(address_el, f"./{{{NS_RAM}}}PostcodeCode")
    line_one = _find_text_optional(address_el, f"./{{{NS_RAM}}}LineOne")
    line_two = _find_text_optional(address_el, f"./{{{NS_RAM}}}LineTwo")
    line_three = _find_text_optional(address_el, f"./{{{NS_RAM}}}LineThree")
    city = _find_text_optional(address_el, f"./{{{NS_RAM}}}CityName")
    country_code = _find_text(address_el, f"./{{{NS_RAM}}}CountryID")
    country_sub = _find_text_optional(
        address_el, f"./{{{NS_RAM}}}CountrySubDivisionName"
    )
    return PostalAddress(
        country_code,
        country_sub,
        post_code,
        city,
        line_one,
        line_two,
        line_three,
    )


def _parse_email(parent: ET.Element) -> str | None:
    id_el = parent.find(
        f"./{{{NS_RAM}}}URIUniversalCommunication/{{{NS_RAM}}}URIID"
    )
    if id_el is None:
        return None
    if id_el.attrib.get("schemeID") != "EM":
        raise InvalidXMLError("Invalid schemeID for email")
    if id_el.text is None:
        raise InvalidXMLError("URIID element has no text")
    return id_el.text


def _parse_tax_simple(
    el: ET.Element,
) -> tuple[TaxCategoryCode, Decimal | None]:
    type_code_s = _find_text(el, f"./{{{NS_RAM}}}TypeCode")
    if type_code_s != "VAT":
        raise InvalidXMLError(f"Invalid tax TypeCode: {type_code_s}")
    category_code_s = _find_text(el, f"./{{{NS_RAM}}}CategoryCode")
    try:
        category_code = TaxCategoryCode(category_code_s)
    except ValueError as exc:
        raise InvalidXMLError(str(exc)) from exc
    rate = _find_percent_optional(el, f"./{{{NS_RAM}}}RateApplicablePercent")
    return category_code, rate


def _parse_tax(el: ET.Element, default_currency: str) -> Tax:
    category_code, rate = _parse_tax_simple(el)
    calculated_amount = _find_amount(
        el, f"./{{{NS_RAM}}}CalculatedAmount", default_currency
    )
    exemption_reason = _find_text_optional(
        el, f"./{{{NS_RAM}}}ExemptionReason"
    )
    basis_amount = _find_amount(
        el, f"./{{{NS_RAM}}}BasisAmount", default_currency
    )

    exemption_reason_code: VATExemptionCode | None = None
    exemption_reason_code_s = _find_text_optional(
        el, f"./{{{NS_RAM}}}ExemptionReasonCode"
    )
    if exemption_reason_code_s is not None:
        try:
            exemption_reason_code = VATExemptionCode(exemption_reason_code_s)
        except ValueError as exc:
            raise InvalidXMLError(str(exc)) from exc

    tax_point_date = _find_date_optional(el, f"./{{{NS_RAM}}}TaxPointDate")

    due_date_type_code_s = _find_text_optional(
        el, f"./{{{NS_RAM}}}DueDateTypeCode"
    )
    due_date_type_code: PaymentTimeCode | None = None
    if due_date_type_code_s is not None:
        try:
            due_date_type_code = PaymentTimeCode(int(due_date_type_code_s))
        except ValueError as exc:
            raise InvalidXMLError(str(exc)) from exc

    return Tax(
        calculated_amount,
        basis_amount,
        rate,
        category_code,
        exemption_reason=exemption_reason,
        exemption_reason_code=exemption_reason_code,
        tax_point_date=tax_point_date,
        due_date_type_code=due_date_type_code,
    )


def _parse_billing_period_optional(
    parent: ET.Element,
) -> tuple[date, date] | None:
    el = parent.find(f"./{{{NS_RAM}}}BillingSpecifiedPeriod")
    if el is None:
        return None
    start_date = _find_date(el, f"./{{{NS_RAM}}}StartDateTime")
    end_date = _find_date(el, f"./{{{NS_RAM}}}EndDateTime")
    return start_date, end_date


def _parse_doc_ref(el: ET.Element) -> DocRef:
    id = _find_text_optional(el, f"./{{{NS_RAM}}}IssuerAssignedID")
    type_code = _find_text(el, f"./{{{NS_RAM}}}TypeCode")
    if type_code != "130":
        raise InvalidXMLError(f"Invalid TypeCode: {type_code}")
    ref_type_code: ReferenceQualifierCode | None = None
    ref_type_code_s = _find_text_optional(
        el, f"./{{{NS_RAM}}}ReferenceTypeCode"
    )
    if ref_type_code_s is not None:
        try:
            ref_type_code = ReferenceQualifierCode(ref_type_code_s)
        except ValueError as exc:
            raise InvalidXMLError(str(exc)) from exc
    return (id, ref_type_code)


def _parse_reference_document(el: ET.Element) -> ReferenceDocument:
    id = _find_text(el, f"./{{{NS_RAM}}}IssuerAssignedID")
    type_code_s = _find_text(el, f"./{{{NS_RAM}}}TypeCode")
    try:
        type_code = DocumentTypeCode(int(type_code_s))
    except ValueError as exc:
        raise InvalidXMLError(str(exc)) from exc
    uri = _find_text_optional(el, f"./{{{NS_RAM}}}URIID")
    name = _find_text_optional(el, f"./{{{NS_RAM}}}Name")
    attachment = _parse_attachment_optional(el)
    ref_type_code: ReferenceQualifierCode | None = None
    ref_type_code_s = _find_text_optional(
        el, f"./{{{NS_RAM}}}ReferenceTypeCode"
    )
    if ref_type_code_s is not None:
        try:
            ref_type_code = ReferenceQualifierCode(ref_type_code_s)
        except ValueError as exc:
            raise InvalidXMLError(str(exc)) from exc

    return ReferenceDocument(
        id, type_code, name, uri, attachment, reference_type_code=ref_type_code
    )


def _parse_attachment_optional(parent: ET.Element) -> Attachment | None:
    el = parent.find(f"./{{{NS_RAM}}}AttachmentBinaryObject")
    if el is None:
        return None
    content = el.text
    mime_type = el.attrib.get("mimeCode")
    filename = el.attrib.get("filename")
    if content is None:
        raise InvalidXMLError("AttachmentBinaryObject element has no text")
    if not mime_type:
        raise InvalidXMLError("AttachmentBinaryObject has no mimeCode")
    if mime_type not in ALLOWED_MIME_TYPES:
        raise InvalidXMLError(f"MIME type not allowed: {mime_type}")
    if not filename:
        raise InvalidXMLError("AttachmentBinaryObject has no filename")
    return b64decode(content), mime_type, filename



from __future__ import annotations

import locale
import re
from decimal import Decimal
from typing import Literal, cast


class Money:
    """An amount of money in a certain currency.

    Initialize with a string with the correct amount of decimal places and
    an ISO 4217 currency code.

    >>> money = Money("33.13", "EUR")
    >>> money.amount
    Decimal('33.13')
    >>> money.currency
    'EUR'

    Alternatively, you can initialize with a Decimal object:

    >>> assert Money(Decimal("33.13"), "EUR") == Money("33.13", "EUR")
    """

    def __init__(self, amount: str | Decimal, currency: str) -> None:
        validate_iso_4217_currency(currency)
        if isinstance(amount, str):
            self.amount = Decimal(amount)
        elif isinstance(amount, Decimal):
            self.amount = amount
        else:
            raise TypeError("Amount must be a str or Decimal")
        self.currency = currency

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Money):
            return NotImplemented

        if (self.amount, self.currency) != (
            value.amount,
            value.currency,
        ):
            return False
        if self.amount.as_tuple().exponent != value.amount.as_tuple().exponent:
            return False
        return True

    def __repr__(self) -> str:
        return f"Money('{str(self.amount)}', {self.currency!r})"

    def __str__(self) -> str:
        conv = locale.localeconv()
        # Check for C locale, in which case locale.currency() raises an error.
        if conv["int_frac_digits"] == 127:  # C locale
            return f"{self.currency} {self.amount}"

        formatted_amount = locale.currency(
            self.amount, symbol=False, grouping=True
        )

        precedes = cast(
            Literal[0, 1],
            conv[self.amount < 0 and "n_cs_precedes" or "p_cs_precedes"],
        )
        separated = cast(
            Literal[0, 1],
            conv[self.amount < 0 and "n_sep_by_space" or "p_sep_by_space"],
        )

        if precedes:
            return self.currency + (separated and " " or "") + formatted_amount
        else:
            currency = (
                self.currency[:-1]
                if self.currency.endswith(" ")
                else self.currency
            )
            return formatted_amount + (separated and " " or "") + currency

    def __mul__(self, other: Decimal) -> Money:
        value = (self.amount * other).quantize(Decimal("1.00"))
        return Money(value, self.currency)

    def __truediv__(self, other: Decimal) -> Money:
        value = (self.amount / other).quantize(Decimal("1.00"))
        return Money(value, self.currency)


_ISO_4217_RE = re.compile(r"^[A-Z]{3}$")


def validate_iso_4217_currency(currency: str) -> None:
    """Validate an ISO 4217 currency code.

    Raise a ValueError if the currency code does not match ISO 4217 format.
    This does not check whether the currency code is actually defined in
    ISO 4217.
    """
    if not _ISO_4217_RE.match(currency):
        raise ValueError(f"Invalid ISO 4217 currency code: {currency}")



"""
Quantities defined in UNECE/CEFACT Trade Facilitation Recommendations
No.20 and No.21.

https://unece.org/trade/uncefact/cl-recommendations
"""

from enum import StrEnum


def N_(s: str) -> str:
    return s


class QuantityCode(StrEnum):
    """
    Selected quantities defined in UNECE/CEFACT Trade Facilitation
    Recommendations No.20 and No.21.
    """

    ONE = "C62"  # aka "unit"
    PIECE = "H87"  # aka "item"
    HOUR = "HUR"
    DAY = "DAY"
    LITER = "LTR"
    CUBIC_METERS = "MTQ"
    KILOGRAM = "KGM"
    METERS = "MTR"
    TON = "TNE"


# Mapping between QuantityCode and its names, either in singular and plural
# forms (as a tuple, localizable) or as a unit symbol (str, not localizable).
QUANTITY_NAMES: dict[QuantityCode, tuple[str, str] | str] = {
    QuantityCode.ONE: (N_("unit"), N_("units")),
    QuantityCode.PIECE: (N_("pc"), N_("pcs")),
    QuantityCode.HOUR: "h",
    QuantityCode.DAY: (N_("day"), N_("days")),
    QuantityCode.LITER: "l",
    QuantityCode.CUBIC_METERS: "m²",
    QuantityCode.KILOGRAM: "kg",
    QuantityCode.METERS: "m",
    QuantityCode.TON: "t",
}



"""
UNTDID type codes, maintained by UN/CEFACT.
"""

from enum import IntEnum, StrEnum


def N_(s: str) -> str:
    return s


ALLOWED_MIME_TYPES = [
    "application/pdf",
    "image/png",
    "image/jpeg",
    "text/csv",
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    "application/vnd.oasis.opendocument.spreadsheet",
]


class IdentifierSchemeCode(StrEnum):
    """Selected identifier scheme codes defined in ISO/IEC 6523."""

    GLN = "0088"  # Global Location Number, aka GS1/EAN location number
    GTIN = "0160"


class DocumentTypeCode(IntEnum):
    """Selected document type codes defined in UNTDID 1001.

    https://service.unece.org/trade/untdid/d99a/uncl/uncl1001.htm
    """

    VALIDATED_PRICED_TENDER = 50
    INVOICING_DATA_SHEET = 130
    PRO_FORMA_INVOICE = 325
    PARTIAL_INVOICE = 326
    INVOICE = 380
    CREDIT_NOTE = 381
    CORRECTION = 384
    PREPAYMENT = 386
    RELATED_DOCUMENT = 916

    @property
    def is_invoice_type(self) -> bool:
        return self in (
            DocumentTypeCode.PRO_FORMA_INVOICE,
            DocumentTypeCode.PARTIAL_INVOICE,
            DocumentTypeCode.INVOICE,
            DocumentTypeCode.CREDIT_NOTE,
            DocumentTypeCode.CORRECTION,
            DocumentTypeCode.PREPAYMENT,
        )

    @property
    def is_supporting_document_type(self) -> bool:
        return self in (
            DocumentTypeCode.VALIDATED_PRICED_TENDER,
            DocumentTypeCode.INVOICING_DATA_SHEET,
            DocumentTypeCode.RELATED_DOCUMENT,
        )


DOCUMENT_TYPE_NAMES = {
    DocumentTypeCode.VALIDATED_PRICED_TENDER: N_("Validated Priced Tender"),
    DocumentTypeCode.INVOICING_DATA_SHEET: N_("Invoicing Data Sheet"),
    DocumentTypeCode.PRO_FORMA_INVOICE: N_("Pro Forma Invoice"),
    DocumentTypeCode.PARTIAL_INVOICE: N_("Partial Invoice"),
    DocumentTypeCode.INVOICE: N_("Invoice"),
    DocumentTypeCode.CREDIT_NOTE: N_("Credit Note"),
    DocumentTypeCode.CORRECTION: N_("Correction"),
    DocumentTypeCode.PREPAYMENT: N_("Prepayment"),
    DocumentTypeCode.RELATED_DOCUMENT: N_("Related Document"),
}


class ReferenceQualifierCode(StrEnum):
    """Selected reference qualifier codes defined in UNTDID 1153.

    https://service.unece.org/trade/untdid/d98b/uncl/uncl1153.htm"""

    PRICE_LIST_VERSION = "PI"


REFERENCE_QUALIFIER_NAMES = {
    ReferenceQualifierCode.PRICE_LIST_VERSION: N_("Price List Version"),
}


class PaymentTimeCode(IntEnum):
    """Selected payment time codes defined in UNTDID 2475.

    https://service.unece.org/trade/untdid/d98b/uncl/uncl2475.htm
    """

    INVOICE_DATE = 5
    DELIVERY_DATE = 29
    PAYMENT_DATE = 72

    @property
    def is_invoice_due_date(self) -> bool:
        return self in (
            PaymentTimeCode.INVOICE_DATE,
            PaymentTimeCode.DELIVERY_DATE,
            PaymentTimeCode.PAYMENT_DATE,
        )


PAYMENT_TIME_CODE_NAMES = {
    PaymentTimeCode.INVOICE_DATE: N_("Invoice Date"),
    PaymentTimeCode.DELIVERY_DATE: N_("Delivery Date"),
    PaymentTimeCode.PAYMENT_DATE: N_("Payment Date"),
}


class TextSubjectCode(StrEnum):
    """Selected text subject codes defined in UNTDID 4451.

    https://service.unece.org/trade/untdid/d96b/uncl/uncl4451.htm
    """

    GENERAL_INFORMATION = "AAI"
    COMMENTS_BY_SELLER = "SUR"
    REGULATORY_INFORMATION = "REG"
    LEGAL_INFORMATION = "ABL"
    TAX_INFORMATION = "TXD"
    CUSTOMS_INFORMATION = "CUS"
    TITLE = "AFM"


TEXT_SUBJECT_CODE_NAMES = {
    TextSubjectCode.GENERAL_INFORMATION: N_("General Information"),
    TextSubjectCode.COMMENTS_BY_SELLER: N_("Comments by Seller"),
    TextSubjectCode.REGULATORY_INFORMATION: N_("Regulatory Information"),
    TextSubjectCode.LEGAL_INFORMATION: N_("Legal Information"),
    TextSubjectCode.TAX_INFORMATION: N_("Tax Information"),
    TextSubjectCode.CUSTOMS_INFORMATION: N_("Customs Information"),
    TextSubjectCode.TITLE: N_("Title"),
}


class PaymentMeansCode(StrEnum):
    """Selected payment means codes defined in UNTDID 4461.

    https://service.unece.org/trade/untdid/d98a/uncl/uncl4461.htm
    """
    NOTDEFINED = "1"
    SPECIES = "10"
    CHECK = "20"
    TRANSFER = "30"
    BANK_PAYMENT = "42"
    CREDIT_CARD = "48"
    DIRECT_DEBIT = "49"
    STANDING_AGREEMENT = "57"
    SEPA_CREDIT_TRANSFER = "58"
    SEPA_DIRECT_DEBIT = "59"
    REPORT = "97"
    INTERIM_AGREEMENT = "ZZZ"


PAYMENT_MEANS_NAMES = {
    PaymentMeansCode.NOTDEFINED: N_("Instrument not defined"),
    PaymentMeansCode.SPECIES: N_("Species"),
    PaymentMeansCode.CHECK: N_("Check"),
    PaymentMeansCode.TRANSFER: N_("Transfer"),
    PaymentMeansCode.BANK_PAYMENT: N_("Bank Payment"),
    PaymentMeansCode.CREDIT_CARD: N_("Credit Card"),
    PaymentMeansCode.DIRECT_DEBIT: N_("Direct Debit"),
    PaymentMeansCode.STANDING_AGREEMENT: N_("Standing Agreement"),
    PaymentMeansCode.SEPA_CREDIT_TRANSFER: N_("SEPA Credit Transfer"),
    PaymentMeansCode.SEPA_DIRECT_DEBIT: N_("SEPA Direct Debit"),
    PaymentMeansCode.REPORT: N_("Report"),
    PaymentMeansCode.INTERIM_AGREEMENT: N_("Interim Agreement"),
}


class AllowanceChargeCode(IntEnum):
    """Selected allowance/charge/discount codes defined in EN 16931.

    Despite the standard saying different, these codes only partially
    correspond to UNTDID 5189.

    https://service.unece.org/trade/untdid/d95b/uncl/uncl5189.htm
    """

    AHEAD_OF_SCHEDULE = 41


class TaxCategoryCode(StrEnum):
    """Duty/tax/fee category codes defined in UNTDID 5305.

    https://service.unece.org/trade/untdid/d98a/uncl/uncl5305.htm
    """

    REVERSE_CHARGE = "AE"  # not in UNTDID 5305, but defined in EN 16931
    EXEMPT = "E"
    FREE_EXPORT = "G"
    INTRA_COMMUNITY_EXEMPT = "K"  # not in UNTDID 5305, but defined in EN 16931
    CANARY_ISLANDS_TAX = "L"  # not in UNTDID 5305, but defined in EN 16931
    CEUTA_MELILLA_TAX = "M"  # not in UNTDID 5305, but defined in EN 16931
    OUT_OF_SCOPE = "O"
    STANDARD_RATE = "S"
    ZERO_RATE = "Z"


class ItemTypeCode(StrEnum):
    """Selected item type codes defined in UNTDID 7143.

    https://service.unece.org/trade/untdid/d96a/uncl/uncl7143.htm
    """

    ISBN = "IB"
    ISSN = "IS"


class SpecialServiceCode(StrEnum):
    """Selected special service (surcharge) codes defined in UNTDID 7161.

    https://service.unece.org/trade/untdid/d98a/uncl/uncl7161.htm
    """

    MATERIAL_SURCHARGE = "MC"


class VATExemptionCode(StrEnum):
    """VAT exemption codes defined by the Connecting Europe Facility (CEF).

    https://ec.europa.eu/digital-building-blocks/sites/display/DIGITAL/Registry+of+supporting+artefacts+to+implement+EN16931
    """



from pathlib import Path
from typing import Final

from ._locale import setup_locale
from .const import (
    URN_BASIC_PROFILE,
    URN_BASIC_WL_PROFILE,
    URN_EN16931_PROFILE,
    URN_EXTENDED_PROFILE,
    URN_MINIMUM_PROFILE,
    URN_XRECHNUNG_PROFILE,
)
from .exc import NoFacturXError
from .model import MinimumInvoice
from .parse import parse_xml
from .pdf_common import FileRelationship
from .pdf_extract import extract_facturx_from_pdf

_ = setup_locale()


def parse_pdf(
    filename: str | Path, *, country: str | None = None
) -> MinimumInvoice:
    """Parse a Factur-X invoice from a PDF file.

    Set the "country" parameter to an ISO 3166-1 alpha-2 country code to
    validate the invoice according to the country-specific rules,
    """
    data, relationship = extract_facturx_from_pdf(filename)
    invoice = parse_xml(data)
    _validate_relationship(invoice.PROFILE_URN, relationship, country=country)
    return invoice


_VALID_RELATIONSHIPS_FRANCE: Final[dict[str, set[FileRelationship]]] = {
    URN_MINIMUM_PROFILE: {
        FileRelationship.DATA,
    },
    URN_BASIC_WL_PROFILE: {
        FileRelationship.DATA,
    },
    URN_BASIC_PROFILE: {
        FileRelationship.ALTERNATIVE,
        FileRelationship.SOURCE,
        FileRelationship.DATA,
    },
    URN_EN16931_PROFILE: {
        FileRelationship.ALTERNATIVE,
        FileRelationship.SOURCE,
        FileRelationship.DATA,
    },
    URN_EXTENDED_PROFILE: {
        FileRelationship.ALTERNATIVE,
        FileRelationship.SOURCE,
        FileRelationship.DATA,
    },
    URN_XRECHNUNG_PROFILE: set(),
}

_VALID_RELATIONSHIPS_GERMANY: Final[dict[str, set[FileRelationship]]] = {
    URN_MINIMUM_PROFILE: {
        FileRelationship.DATA,
    },
    URN_BASIC_WL_PROFILE: {
        FileRelationship.DATA,
    },
    URN_BASIC_PROFILE: {
        FileRelationship.ALTERNATIVE,
    },
    URN_EN16931_PROFILE: {
        FileRelationship.ALTERNATIVE,
    },
    URN_EXTENDED_PROFILE: {
        FileRelationship.ALTERNATIVE,
    },
    URN_XRECHNUNG_PROFILE: {
        FileRelationship.ALTERNATIVE,
    },
}


def _validate_relationship(
    profile: str,
    relationship: FileRelationship | None,
    *,
    country: str | None = None,
) -> None:
    """Validate that the PDF file relationship matches the Factur-X invoice.

    See Factur-X 1.0.07, section 6.2.2.
    """
    if relationship is None:
        return

    match country:
        case "fr":
            rel_map = _VALID_RELATIONSHIPS_FRANCE
        case "de":
            rel_map = _VALID_RELATIONSHIPS_GERMANY
        case _:
            return

    if relationship not in rel_map[profile]:
        raise NoFacturXError(
            _("Invalid relationship for Factur-X Minimum invoice")
        )
